import json
import platform
from os import path
from typing import Optional

import click

from toot import __version__, api, config, settings
from toot.cli import cli
from toot.entities import Instance, from_dict
from toot.output import bold, yellow
from toot.utils import get_distro_name, get_version

DIAG_DEPENDENCIES = [
    "beautifulsoup4",
    "click",
    "pillow",
    "requests",
    "setuptools",
    "term-image",
    "tomlkit",
    "typing-extensions",
    "urwid",
    "urwidgets",
    "wcwidth",
]


@cli.command()
@click.option(
    "-f",
    "--files",
    is_flag=True,
    help="Print contents of the config and settings files in diagnostic output",
)
@click.option(
    "-s",
    "--server",
    is_flag=True,
    help="Print information about the curren server in diagnostic output",
)
def diag(files: bool, server: bool):
    """Display useful information for diagnosing problems"""
    print_diag(files, server)


def print_diag(files: bool, server: bool):
    instance: Optional[Instance] = None
    if server:
        _, app = config.get_active_user_app()
        if app:
            response = api.get_instance(app.base_url)
            instance = from_dict(Instance, response.json())

    click.echo("## Toot Diagnostics")
    print_environment()
    print_dependencies()
    print_instance(instance)
    print_settings(files)
    print_config(files)


def print_environment():
    click.echo()
    click.echo(f"toot {__version__}")
    click.echo(f"Python {platform.python_version()}")
    click.echo(platform.platform())

    distro = get_distro_name()
    if distro:
        click.echo(distro)


def print_dependencies():
    click.echo()
    click.secho(bold("Dependencies:"))
    for dep in DIAG_DEPENDENCIES:
        version = get_version(dep) or yellow("not installed")
        click.echo(f" * {dep}: {version}")


def print_instance(instance: Optional[Instance]):
    if instance:
        click.echo()
        click.echo(bold("Server:"))
        click.echo(instance.title)
        click.echo(instance.uri)
        click.echo(f"version {instance.version}")


def print_settings(include_files: bool):
    click.echo()
    settings_path = settings.get_settings_path()
    if path.exists(settings_path):
        click.echo(f"Settings file: {settings_path}")
        if include_files:
            with open(settings_path, "r") as f:
                click.echo("\n```toml")
                click.echo(f.read().strip())
                click.echo("```\n")
    else:
        click.echo(f'Settings file: {yellow("not found")}')


def print_config(include_files: bool):
    click.echo()
    config_path = config.get_config_file_path()
    if path.exists(config_path):
        click.echo(f"Config file: {config_path}")
        if include_files:
            content = _get_anonymized_config(config_path)
            click.echo("\n```json")
            click.echo(json.dumps(content, indent=4))
            click.echo("```\n")
    else:
        click.echo(f'Config file: {yellow("not found")}')


def _get_anonymized_config(config_path):
    with open(config_path, "r") as f:
        content = json.load(f)

        for app in content.get("apps", {}).values():
            app["client_id"] = "*****"
            app["client_secret"] = "*****"

        for user in content.get("users", {}).values():
            user["access_token"] = "*****"

        return content



from urwid.command_map import command_map
from urwid.command_map import CURSOR_UP, CURSOR_DOWN, CURSOR_LEFT, CURSOR_RIGHT

# Add movement using h/j/k/l to default command map
command_map._command.update({
    'k': CURSOR_UP,
    'j': CURSOR_DOWN,
    'h': CURSOR_LEFT,
    'l': CURSOR_RIGHT,
})



import mimetypes
import re
import uuid

from os import path
from requests import Response
from typing import BinaryIO, List, Optional
from urllib.parse import urlparse, urlencode, quote

from toot import App, User, http, CLIENT_NAME, CLIENT_WEBSITE
from toot.exceptions import ConsoleError
from toot.utils import drop_empty_values, str_bool, str_bool_nullable


SCOPES = 'read write follow'


def find_account(app, user, account_name):
    if not account_name:
        raise ConsoleError("Empty account name given")

    normalized_name = account_name.lstrip("@").lower()

    # Strip @<instance_name> from accounts on the local instance. The `acct`
    # field in account object contains the qualified name for users of other
    # instances, but only the username for users of the local instance. This is
    # required in order to match the account name below.
    if "@" in normalized_name:
        [username, instance] = normalized_name.split("@", maxsplit=1)
        if instance == app.instance:
            normalized_name = username

    response = search(app, user, account_name, type="accounts", resolve=True)
    for account in response.json()["accounts"]:
        if account["acct"].lower() == normalized_name:
            return account

    raise ConsoleError("Account not found")


def lookup(app, user, acct):
    return http.get(app, user, "/api/v1/accounts/lookup", {"acct": acct})


def _account_action(app, user, account, action) -> Response:
    url = f"/api/v1/accounts/{account}/{action}"
    return http.post(app, user, url)


def _status_action(app, user, status_id, action, data=None) -> Response:
    status_id = _resolve_status_id(app, user, status_id)
    url = f"/api/v1/statuses/{status_id}/{action}"
    return http.post(app, user, url, data=data)


def _resolve_status_id(app, user, id_or_url) -> str:
    """
    If given an URL instead of status ID, attempt to resolve the status ID.

    TODO: Not 100% sure this is the correct way of doing this, but it seems to
    work for all test cases I've thrown at it. So leaving it undocumented until
    we're happy it works.
    """
    if re.match(r"^https?://", id_or_url):
        response = search(app, user, id_or_url, resolve=True, type="statuses")
        statuses = response.json().get("statuses")

        if not statuses:
            raise ConsoleError(f"Cannot find status matching URL {id_or_url}")

        if len(statuses) > 1:
            raise ConsoleError(f"Found multiple statuses mathcing URL {id_or_url}")

        return statuses[0]["id"]

    return id_or_url


def _tag_action(app, user, tag_name, action) -> Response:
    url = f"/api/v1/tags/{tag_name}/{action}"
    return http.post(app, user, url)


def create_app(base_url):
    url = f"{base_url}/api/v1/apps"

    json = {
        'client_name': CLIENT_NAME,
        'redirect_uris': 'urn:ietf:wg:oauth:2.0:oob',
        'scopes': SCOPES,
        'website': CLIENT_WEBSITE,
    }

    return http.anon_post(url, json=json).json()


def get_muted_accounts(app, user):
    return http.get(app, user, "/api/v1/mutes").json()


def get_blocked_accounts(app, user):
    return http.get(app, user, "/api/v1/blocks").json()


def register_account(app, username, email, password, locale="en", agreement=True):
    """
    Register an account
    https://docs.joinmastodon.org/methods/accounts/#create
    """
    token = fetch_app_token(app)["access_token"]
    url = f"{app.base_url}/api/v1/accounts"
    headers = {"Authorization": f"Bearer {token}"}

    json = {
        "username": username,
        "email": email,
        "password": password,
        "agreement": agreement,
        "locale": locale
    }

    return http.anon_post(url, json=json, headers=headers).json()


def update_account(
    app,
    user,
    display_name=None,
    note=None,
    avatar=None,
    header=None,
    bot=None,
    discoverable=None,
    locked=None,
    privacy=None,
    sensitive=None,
    language=None
):
    """
    Update account credentials
    https://docs.joinmastodon.org/methods/accounts/#update_credentials
    """
    files = drop_empty_values({"avatar": avatar, "header": header})

    data = drop_empty_values({
        "bot": str_bool_nullable(bot),
        "discoverable": str_bool_nullable(discoverable),
        "display_name": display_name,
        "locked": str_bool_nullable(locked),
        "note": note,
        "source[language]": language,
        "source[privacy]": privacy,
        "source[sensitive]": str_bool_nullable(sensitive),
    })

    return http.patch(app, user, "/api/v1/accounts/update_credentials", files=files, data=data)


def fetch_app_token(app):
    json = {
        "client_id": app.client_id,
        "client_secret": app.client_secret,
        "grant_type": "client_credentials",
        "redirect_uri": "urn:ietf:wg:oauth:2.0:oob",
        "scope": "read write"
    }

    return http.anon_post(f"{app.base_url}/oauth/token", json=json).json()


def login(app: App, username: str, password: str):
    url = app.base_url + '/oauth/token'

    data = {
        'grant_type': 'password',
        'client_id': app.client_id,
        'client_secret': app.client_secret,
        'username': username,
        'password': password,
        'scope': SCOPES,
    }

    return http.anon_post(url, data=data).json()


def get_browser_login_url(app: App) -> str:
    """Returns the URL for manual log in via browser"""
    return "{}/oauth/authorize/?{}".format(app.base_url, urlencode({
        "response_type": "code",
        "redirect_uri": "urn:ietf:wg:oauth:2.0:oob",
        "scope": SCOPES,
        "client_id": app.client_id,
    }))


def request_access_token(app: App, authorization_code: str):
    url = app.base_url + '/oauth/token'

    data = {
        'grant_type': 'authorization_code',
        'client_id': app.client_id,
        'client_secret': app.client_secret,
        'code': authorization_code,
        'redirect_uri': 'urn:ietf:wg:oauth:2.0:oob',
    }

    return http.anon_post(url, data=data, allow_redirects=False).json()


def post_status(
    app,
    user,
    status,
    visibility=None,
    media_ids=None,
    sensitive=False,
    spoiler_text=None,
    in_reply_to_id=None,
    language=None,
    scheduled_at=None,
    content_type=None,
    poll_options=None,
    poll_expires_in=None,
    poll_multiple=None,
    poll_hide_totals=None,
) -> Response:
    """
    Publish a new status.
    https://docs.joinmastodon.org/methods/statuses/#create
    """

    # Idempotency key assures the same status is not posted multiple times
    # if the request is retried.
    headers = {"Idempotency-Key": uuid.uuid4().hex}

    # Strip keys for which value is None
    # Sending null values doesn't bother Mastodon, but it breaks Pleroma
    data = drop_empty_values({
        'status': status,
        'media_ids': media_ids,
        'visibility': visibility,
        'sensitive': sensitive,
        'in_reply_to_id': in_reply_to_id,
        'language': language,
        'scheduled_at': scheduled_at,
        'content_type': content_type,
        'spoiler_text': spoiler_text,
    })

    if poll_options:
        data["poll"] = {
            "options": poll_options,
            "expires_in": poll_expires_in,
            "multiple": poll_multiple,
            "hide_totals": poll_hide_totals,
        }

    return http.post(app, user, '/api/v1/statuses', json=data, headers=headers)


def edit_status(
    app,
    user,
    id,
    status,
    visibility='public',
    media_ids=None,
    sensitive=False,
    spoiler_text=None,
    in_reply_to_id=None,
    language=None,
    content_type=None,
    poll_options=None,
    poll_expires_in=None,
    poll_multiple=None,
    poll_hide_totals=None,
) -> Response:
    """
    Edit an existing status
    https://docs.joinmastodon.org/methods/statuses/#edit
    """

    # Strip keys for which value is None
    # Sending null values doesn't bother Mastodon, but it breaks Pleroma
    data = drop_empty_values({
        'status': status,
        'media_ids': media_ids,
        'visibility': visibility,
        'sensitive': sensitive,
        'in_reply_to_id': in_reply_to_id,
        'language': language,
        'content_type': content_type,
        'spoiler_text': spoiler_text,
    })

    if poll_options:
        data["poll"] = {
            "options": poll_options,
            "expires_in": poll_expires_in,
            "multiple": poll_multiple,
            "hide_totals": poll_hide_totals,
        }

    return http.put(app, user, f"/api/v1/statuses/{id}", json=data)


def fetch_status(app, user, id):
    """
    Fetch a single status
    https://docs.joinmastodon.org/methods/statuses/#get
    """
    return http.get(app, user, f"/api/v1/statuses/{id}")


def fetch_status_source(app, user, id):
    """
    Fetch the source (original text) for a single status.
    This only works on local toots.
    https://docs.joinmastodon.org/methods/statuses/#source
    """
    return http.get(app, user, f"/api/v1/statuses/{id}/source")


def scheduled_statuses(app, user):
    """
    List scheduled statuses
    https://docs.joinmastodon.org/methods/scheduled_statuses/#get
    """
    return http.get(app, user, "/api/v1/scheduled_statuses").json()


def delete_status(app, user, status_id):
    """
    Deletes a status with given ID.
    https://github.com/tootsuite/documentation/blob/master/Using-the-API/API.md#deleting-a-status
    """
    return http.delete(app, user, f"/api/v1/statuses/{status_id}")


def favourite(app, user, status_id):
    return _status_action(app, user, status_id, 'favourite')


def unfavourite(app, user, status_id):
    return _status_action(app, user, status_id, 'unfavourite')


def reblog(app, user, status_id, visibility="public"):
    return _status_action(app, user, status_id, 'reblog', data={"visibility": visibility})


def unreblog(app, user, status_id):
    return _status_action(app, user, status_id, 'unreblog')


def pin(app, user, status_id):
    return _status_action(app, user, status_id, 'pin')


def unpin(app, user, status_id):
    return _status_action(app, user, status_id, 'unpin')


def bookmark(app, user, status_id):
    return _status_action(app, user, status_id, 'bookmark')


def unbookmark(app, user, status_id):
    return _status_action(app, user, status_id, 'unbookmark')


def translate(app, user, status_id):
    return _status_action(app, user, status_id, 'translate')


def context(app, user, status_id) -> Response:
    url = f"/api/v1/statuses/{status_id}/context"
    return http.get(app, user, url)


def reblogged_by(app, user, status_id) -> Response:
    url = f"/api/v1/statuses/{status_id}/reblogged_by"
    return http.get(app, user, url)


def get_timeline_generator(
    app: Optional[App],
    user: Optional[User],
    account: Optional[str] = None,
    list_id: Optional[str] = None,
    tag: Optional[str] = None,
    local: bool = False,
    public: bool = False,
    limit: int = 20,  # TODO
):
    if public:
        return public_timeline_generator(app, user, local=local, limit=limit)
    elif tag:
        return tag_timeline_generator(app, user, tag, local=local, limit=limit)
    elif account:
        return account_timeline_generator(app, user, account, limit=limit)
    elif list_id:
        return timeline_list_generator(app, user, list_id, limit=limit)
    else:
        return home_timeline_generator(app, user, limit=limit)


def _get_next_path(headers):
    """Given timeline response headers, returns the path to the next batch"""
    links = headers.get('Link', '')
    matches = re.match('<([^>]+)>; rel="next"', links)
    if matches:
        parsed = urlparse(matches.group(1))
        return "?".join([parsed.path, parsed.query])


def _get_next_url(headers) -> Optional[str]:
    """Given timeline response headers, returns the url to the next batch"""
    links = headers.get('Link', '')
    match = re.match('<([^>]+)>; rel="next"', links)
    if match:
        return match.group(1)


def _timeline_generator(app, user, path, params=None):
    if params:
        path += f"?{urlencode(params)}"

    while path:
        response = http.get(app, user, path)
        yield response.json()
        path = _get_next_path(response.headers)


def _notification_timeline_generator(app, user, path, params=None):
    for batch in _timeline_generator(app, user, path, params):
        yield [n["status"] for n in batch if n.get("status")]


def _conversation_timeline_generator(app, user, path, params=None):
    for batch in _timeline_generator(app, user, path, params):
        yield [c["last_status"] for c in batch if c.get("last_status")]


def home_timeline_generator(app, user, limit=20):
    path = "/api/v1/timelines/home"
    params = {"limit": limit}
    return _timeline_generator(app, user, path, params)


def public_timeline_generator(app, user, local=False, limit=20):
    path = '/api/v1/timelines/public'
    params = {'local': str_bool(local), 'limit': limit}
    return _timeline_generator(app, user, path, params)


def tag_timeline_generator(app, user, hashtag, local=False, limit=20):
    path = f"/api/v1/timelines/tag/{quote(hashtag)}"
    params = {'local': str_bool(local), 'limit': limit}
    return _timeline_generator(app, user, path, params)


def bookmark_timeline_generator(app, user, limit=20):
    path = '/api/v1/bookmarks'
    params = {'limit': limit}
    return _timeline_generator(app, user, path, params)


def notification_timeline_generator(app, user, limit=20):
    # exclude all but mentions and statuses
    exclude_types = ["follow", "favourite", "reblog", "poll", "follow_request"]
    params = {"exclude_types[]": exclude_types, "limit": limit}
    return _notification_timeline_generator(app, user, "/api/v1/notifications", params)


def conversation_timeline_generator(app, user, limit=20):
    path = "/api/v1/conversations"
    params = {"limit": limit}
    return _conversation_timeline_generator(app, user, path, params)


def account_timeline_generator(app, user, account_name: str, replies=False, reblogs=False, limit=20):
    account = find_account(app, user, account_name)
    path = f"/api/v1/accounts/{account['id']}/statuses"
    params = {"limit": limit, "exclude_replies": not replies, "exclude_reblogs": not reblogs}
    return _timeline_generator(app, user, path, params)


def account_timeline_generator_by_id(app, user, account_id: str, replies=False, reblogs=False, limit=20):
    path = f"/api/v1/accounts/{account_id}/statuses"
    params = {"limit": limit, "exclude_replies": not replies, "exclude_reblogs": not reblogs}
    return _timeline_generator(app, user, path, params)


def timeline_list_generator(app, user, list_id, limit=20):
    path = f"/api/v1/timelines/list/{list_id}"
    return _timeline_generator(app, user, path, {'limit': limit})


def _anon_timeline_generator(url, params=None):
    if params:
        url += f"?{urlencode(params)}"

    while url:
        response = http.anon_get(url)
        yield response.json()
        url = _get_next_url(response.headers)


def anon_public_timeline_generator(base_url, local=False, limit=20):
    params = {"local": str_bool(local), "limit": limit}
    url = f"{base_url}/api/v1/timelines/public"
    return _anon_timeline_generator(url, params)


def anon_tag_timeline_generator(base_url, hashtag, local=False, limit=20):
    query = {"local": str_bool(local), "limit": limit}
    url = f"{base_url}/api/v1/timelines/tag/{quote(hashtag)}"
    return _anon_timeline_generator(url, query)


def get_media(app: App, user: User, id: str):
    return http.get(app, user, f"/api/v1/media/{id}").json()


def upload_media(
    app: App,
    user: User,
    media: BinaryIO,
    description: Optional[str] = None,
    thumbnail: Optional[BinaryIO] = None,
):
    data = drop_empty_values({"description": description})

    # NB: Documentation says that "file" should provide a mime-type which we
    # don't do currently, but it works.
    files = drop_empty_values({
        "file": media,
        "thumbnail": _add_mime_type(thumbnail)
    })

    return http.post(app, user, "/api/v2/media", data=data, files=files)


def _add_mime_type(file):
    if file is None:
        return None

    # TODO: mimetypes uses the file extension to guess the mime type which is
    # not always good enough (e.g. files without extension). python-magic could
    # be used instead but it requires adding it as a dependency.
    mime_type = mimetypes.guess_type(file.name)

    if not mime_type:
        raise ConsoleError(f"Unable guess mime type of '{file.name}'. "
                           "Ensure the file has the desired extension.")

    filename = path.basename(file.name)
    return (filename, file, mime_type)


def search(
    app,
    user,
    query,
    resolve=False,
    type=None,
    offset=None,
    limit=None,
    min_id=None,
    max_id=None,
):
    """
    Perform a search.
    https://docs.joinmastodon.org/methods/search/#v2
    """
    params = drop_empty_values({
        "q": query,
        "resolve": str_bool(resolve),
        "type": type,
        "offset": offset,
        "limit": limit,
        "min_id": min_id,
        "max_id": max_id,
    })

    return http.get(app, user, "/api/v2/search", params)


def follow(app, user, account):
    return _account_action(app, user, account, 'follow')


def unfollow(app, user, account):
    return _account_action(app, user, account, 'unfollow')


def follow_tag(app, user, tag_name) -> Response:
    return _tag_action(app, user, tag_name, 'follow')


def unfollow_tag(app, user, tag_name) -> Response:
    return _tag_action(app, user, tag_name, 'unfollow')


def _get_response_list(app, user, path):
    items = []
    while path:
        response = http.get(app, user, path)
        items += response.json()
        path = _get_next_path(response.headers)
    return items


def following(app, user, account):
    path = f"/api/v1/accounts/{account}/following"
    return _get_response_list(app, user, path)


def followers(app, user, account):
    path = f"/api/v1/accounts/{account}/followers"
    return _get_response_list(app, user, path)


def followed_tags(app, user):
    path = '/api/v1/followed_tags'
    return _get_response_list(app, user, path)


def featured_tags(app, user):
    return http.get(app, user, "/api/v1/featured_tags")


def feature_tag(app, user, tag: str) -> Response:
    return http.post(app, user, "/api/v1/featured_tags", data={"name": tag})


def unfeature_tag(app, user, tag_id: str) -> Response:
    return http.delete(app, user, f"/api/v1/featured_tags/{tag_id}")


def find_tag(app, user, tag) -> Optional[dict]:
    """Find a hashtag by tag name or ID"""
    tag = tag.lstrip("#")
    results = search(app, user, tag, type="hashtags").json()

    return next(
        (
            t for t in results["hashtags"]
            if t["name"].lower() == tag.lstrip("#").lower() or t["id"] == tag
        ),
        None
    )


def find_featured_tag(app, user, tag) -> Optional[dict]:
    """Find a featured tag by tag name or ID"""
    return next(
        (
            t for t in featured_tags(app, user).json()
            if t["name"].lower() == tag.lstrip("#").lower() or t["id"] == tag
        ),
        None
    )


def whois(app, user, account):
    return http.get(app, user, f'/api/v1/accounts/{account}').json()


def vote(app, user, poll_id, choices: List[int]):
    url = f"/api/v1/polls/{poll_id}/votes"
    json = {'choices': choices}
    return http.post(app, user, url, json=json).json()


def get_relationship(app, user, account):
    params = {"id[]": account}
    return http.get(app, user, '/api/v1/accounts/relationships', params).json()[0]


def mute(app, user, account):
    return _account_action(app, user, account, 'mute')


def unmute(app, user, account):
    return _account_action(app, user, account, 'unmute')


def muted(app, user):
    return _get_response_list(app, user, "/api/v1/mutes")


def block(app, user, account):
    return _account_action(app, user, account, 'block')


def unblock(app, user, account):
    return _account_action(app, user, account, 'unblock')


def blocked(app, user):
    return _get_response_list(app, user, "/api/v1/blocks")


def verify_credentials(app, user) -> Response:
    return http.get(app, user, '/api/v1/accounts/verify_credentials')


def get_notifications(app, user, types=[], exclude_types=[], limit=20) -> Response:
    params = {"types[]": types, "exclude_types[]": exclude_types, "limit": limit}
    return http.get(app, user, '/api/v1/notifications', params)


def clear_notifications(app, user):
    http.post(app, user, '/api/v1/notifications/clear')


def get_instance(base_url: str) -> Response:
    url = f"{base_url}/api/v1/instance"
    return http.anon_get(url)


def get_preferences(app, user) -> Response:
    return http.get(app, user, '/api/v1/preferences')


def get_lists(app, user):
    return http.get(app, user, "/api/v1/lists").json()


def get_list_accounts(app, user, list_id):
    path = f"/api/v1/lists/{list_id}/accounts"
    return _get_response_list(app, user, path)


def create_list(app, user, title, replies_policy="none"):
    url = "/api/v1/lists"
    json = {'title': title}
    if replies_policy:
        json['replies_policy'] = replies_policy
    return http.post(app, user, url, json=json)


def delete_list(app, user, id):
    return http.delete(app, user, f"/api/v1/lists/{id}")


def add_accounts_to_list(app, user, list_id, account_ids):
    url = f"/api/v1/lists/{list_id}/accounts"
    json = {'account_ids': account_ids}
    return http.post(app, user, url, json=json)


def remove_accounts_from_list(app, user, list_id, account_ids):
    url = f"/api/v1/lists/{list_id}/accounts"
    json = {'account_ids': account_ids}
    return http.delete(app, user, url, json=json)



import click
import logging
import os
import sys
import typing as t

from click.shell_completion import CompletionItem
from click.types import StringParamType
from functools import wraps

from toot import App, User, config, __version__
from toot.output import print_warning
from toot.settings import get_settings

if t.TYPE_CHECKING:
    import typing_extensions as te
    P = te.ParamSpec("P")

R = t.TypeVar("R")
T = t.TypeVar("T")


PRIVACY_CHOICES = ["public", "unlisted", "private"]
VISIBILITY_CHOICES = ["public", "unlisted", "private", "direct"]
IMAGE_FORMAT_CHOICES = ["block", "iterm", "kitty"]
NOTIFICATION_TYPE_CHOICES = [
    "mention",
    "status",
    "reblog",
    "follow",
    "follow_request",
    "favourite",
    "poll",
    "update",
    "admin.sign_up",
    "admin.report",
]
TUI_COLORS = {
    "1": 1,
    "16": 16,
    "88": 88,
    "256": 256,
    "16777216": 16777216,
    "24bit": 16777216,
}
TUI_COLORS_CHOICES = list(TUI_COLORS.keys())
TUI_COLORS_VALUES = list(TUI_COLORS.values())

DURATION_EXAMPLES = """e.g. "1 day", "2 hours 30 minutes", "5 minutes 30
seconds" or any combination of above. Shorthand: "1d", "2h30m", "5m30s\""""


def get_default_visibility() -> str:
    return os.getenv("TOOT_POST_VISIBILITY", "public")


def get_default_map():
    settings = get_settings()
    common = settings.get("common", {})
    commands = settings.get("commands", {})

    # TODO: remove in version 1.0
    tui_old = settings.get("tui", {}).copy()
    if "palette" in tui_old:
        del tui_old["palette"]
    if tui_old:
        # TODO: don't show the warning for [toot.palette]
        print_warning("Settings section [tui] has been deprecated in favour of [commands.tui].")
        tui_new = commands.get("tui", {})
        commands["tui"] = {**tui_old, **tui_new}

    return {**common, **commands}


# Tweak the Click context
# https://click.palletsprojects.com/en/8.1.x/api/#context
CONTEXT = dict(
    # Enable using environment variables to set options
    auto_envvar_prefix="TOOT",
    # Add shorthand -h for invoking help
    help_option_names=["-h", "--help"],
    # Always show default values for options
    show_default=True,
    # Load command defaults from settings
    default_map=get_default_map(),
)


class Context(t.NamedTuple):
    app: t.Optional[App]
    user: t.Optional[User] = None
    color: bool = False
    debug: bool = False


class TootObj(t.NamedTuple):
    """Data to add to Click context"""
    color: bool = True
    debug: bool = False
    as_user: t.Optional[str] = None
    # Pass a context for testing purposes
    test_ctx: t.Optional[Context] = None


class AccountParamType(StringParamType):
    """Custom type to add shell completion for account names"""
    name = "account"

    def shell_complete(self, ctx, param, incomplete: str):
        users = config.load_config()["users"].keys()
        return [
            CompletionItem(u)
            for u in users
            if u.lower().startswith(incomplete.lower())
        ]


class InstanceParamType(StringParamType):
    """Custom type to add shell completion for instance domains"""
    name = "instance"

    def shell_complete(self, ctx, param, incomplete: str):
        apps = config.load_config()["apps"]

        return [
            CompletionItem(i)
            for i in apps.keys()
            if i.lower().startswith(incomplete.lower())
        ]


def pass_context(f: "t.Callable[te.Concatenate[Context, P], R]") -> "t.Callable[P, R]":
    """Pass the toot Context as first argument."""
    @wraps(f)
    def wrapped(*args: "P.args", **kwargs: "P.kwargs") -> R:
        return f(get_context(), *args, **kwargs)

    return wrapped


def get_context() -> Context:
    click_context = click.get_current_context()
    obj: TootObj = click_context.obj

    # This is used to pass a context for testing, not used in normal usage
    if obj.test_ctx:
        return obj.test_ctx

    if obj.as_user:
        user, app = config.get_user_app(obj.as_user)
        if not user or not app:
            raise click.ClickException(f"Account '{obj.as_user}' not found. Run `toot auth` to see available accounts.")
    else:
        user, app = config.get_active_user_app()
        if not user or not app:
            raise click.ClickException("This command requires you to be logged in.")

    return Context(app, user, obj.color, obj.debug)


json_option = click.option(
    "--json",
    is_flag=True,
    default=False,
    help="Print data as JSON rather than human readable text"
)


@click.group(context_settings=CONTEXT)
@click.option("-w", "--max-width", type=int, default=80, help="Maximum width for content rendered by toot")
@click.option("--debug/--no-debug", default=False, help="Log debug info to stderr")
@click.option("--color/--no-color", default=sys.stdout.isatty(), help="Use ANSI color in output")
@click.option("--as", "as_user", type=AccountParamType(), help="The account to use, overrides the active account.")
@click.version_option(__version__, message="%(prog)s v%(version)s")
@click.pass_context
def cli(ctx: click.Context, max_width: int, color: bool, debug: bool, as_user: str):
    """Toot is a Mastodon CLI"""
    ctx.obj = TootObj(color, debug, as_user)
    ctx.color = color
    ctx.max_content_width = max_width

    if debug:
        logging.basicConfig(level=logging.DEBUG)


from toot.cli import accounts  # noqa
from toot.cli import auth  # noqa
from toot.cli import diag  # noqa
from toot.cli import lists  # noqa
from toot.cli import post  # noqa
from toot.cli import read  # noqa
from toot.cli import statuses  # noqa
from toot.cli import tags  # noqa
from toot.cli import timelines  # noqa
from toot.cli import timelines_v2  # noqa
from toot.cli import tui  # noqa



import webbrowser

import click

from toot import api, config
from toot.auth import get_or_create_app, login_auth_code, login_username_password
from toot.cli import AccountParamType, cli
from toot.cli.diag import print_diag
from toot.cli.validators import validate_instance
from toot.output import print_warning

instance_option = click.option(
    "--instance", "-i", "base_url",
    prompt="Enter instance URL",
    default="https://mastodon.social",
    callback=validate_instance,
    help="""Domain or base URL of the instance to log into,
            e.g. 'mastodon.social' or 'https://mastodon.social'""",
)


@cli.command()
def auth():
    """Show logged in accounts and instances"""
    config_data = config.load_config()

    if not config_data["users"]:
        click.echo("You are not logged in to any accounts")
        return

    active_user = config_data["active_user"]

    click.echo("Authenticated accounts:")
    for uid, u in config_data["users"].items():
        active_label = "ACTIVE" if active_user == uid else ""
        uid = click.style(uid, fg="green")
        active_label = click.style(active_label, fg="yellow")
        click.echo(f"* {uid} {active_label}")

    path = config.get_config_file_path()
    path = click.style(path, "blue")
    click.echo(f"\nAuth tokens are stored in: {path}")


@cli.command(hidden=True)
def env():
    """Deprecated in favour of 'diag'"""
    print_warning("`toot env` is deprecated in favour of `toot diag`")
    click.echo()
    print_diag(False, False)


@cli.command(name="login_cli")
@instance_option
@click.option("--email", "-e", help="Email address to log in with", prompt=True)
@click.option("--password", "-p", hidden=True, prompt=True, hide_input=True)
def login_cli(base_url: str, email: str, password: str):
    """
    Log into an instance from the console (not recommended)

    Does NOT support two factor authentication, may not work on instances
    other than Mastodon, mostly useful for scripting.
    """
    app = get_or_create_app(base_url)
    login_username_password(app, email, password)

    click.secho("✓ Successfully logged in.", fg="green")
    click.echo("Access token saved to config at: ", nl=False)
    click.secho(config.get_config_file_path(), fg="green")


LOGIN_EXPLANATION = """This authentication method requires you to log into your
Mastodon instance in your browser, where you will be asked to authorize toot to
access your account. When you do, you will be given an authorization code which
you need to paste here.""".replace("\n", " ")


@cli.command()
@instance_option
def login(base_url: str):
    """Log into an instance using your browser (recommended)"""
    app = get_or_create_app(base_url)
    url = api.get_browser_login_url(app)

    click.echo(click.wrap_text(LOGIN_EXPLANATION))
    click.echo("\nLogin URL:")
    click.echo(url)

    yesno = click.prompt("Open link in default browser? [Y/n]", default="Y", show_default=False)
    if not yesno or yesno.lower() == 'y':
        webbrowser.open(url)

    authorization_code = ""
    while not authorization_code:
        authorization_code = click.prompt("Authorization code")

    login_auth_code(app, authorization_code)

    click.echo()
    click.secho("✓ Successfully logged in.", fg="green")


@cli.command()
@click.argument("account", type=AccountParamType(), required=False)
def logout(account: str):
    """Log out of ACCOUNT, delete stored access keys"""
    accounts = _get_accounts_list()

    if not account:
        raise click.ClickException(f"Specify account to log out:\n{accounts}")

    user = config.load_user(account)

    if not user:
        raise click.ClickException(f"Account not found. Logged in accounts:\n{accounts}")

    config.delete_user(user)
    click.secho(f"✓ Account {account} logged out", fg="green")


@cli.command()
@click.argument("account", type=AccountParamType(), required=False)
def activate(account: str):
    """Switch to logged in ACCOUNT."""
    accounts = _get_accounts_list()

    if not account:
        raise click.ClickException(f"Specify account to activate:\n{accounts}")

    user = config.load_user(account)

    if not user:
        raise click.ClickException(f"Account not found. Logged in accounts:\n{accounts}")

    config.activate_user(user)
    click.secho(f"✓ Account {account} activated", fg="green")


def _get_accounts_list() -> str:
    accounts = config.load_config()["users"].keys()
    if not accounts:
        raise click.ClickException("You're not logged into any accounts")
    return "\n".join([f"* {acct}" for acct in accounts])



import urwid
import math
import requests
import warnings

# If term_image is loaded use their screen implementation which handles images
try:
    from term_image.widget import UrwidImageScreen, UrwidImage
    from term_image.image import BaseImage, KittyImage, ITerm2Image, BlockImage
    from term_image import disable_queries  # prevent phantom keystrokes
    from PIL import Image, ImageDraw

    _IMAGE_PIXEL_FORMATS = frozenset({'kitty', 'iterm'})
    _ImageCls = None

    TuiScreen = UrwidImageScreen
    disable_queries()

    def image_support_enabled():
        return True

    def can_render_pixels(image_format):
        return image_format in _IMAGE_PIXEL_FORMATS

    def get_base_image(image, image_format, colors) -> BaseImage:
        # we don't autodetect kitty, iterm; we choose based on option switches

        global _ImageCls

        if not _ImageCls:
            _ImageCls = (
                KittyImage
                if image_format == 'kitty'
                else ITerm2Image
                if image_format == 'iterm'
                else BlockImage
            )
            _ImageCls.forced_support = True
            if colors == 256 and not can_render_pixels(image_format):
                _ImageCls.set_render_method("INDEXED")

        return _ImageCls(image)

    def resize_image(basewidth: int, baseheight: int, img: Image.Image) -> Image.Image:
        if baseheight and not basewidth:
            hpercent = baseheight / float(img.size[1])
            width = math.ceil(img.size[0] * hpercent)
            img = img.resize((width, baseheight), Image.Resampling.LANCZOS)
        elif basewidth and not baseheight:
            wpercent = (basewidth / float(img.size[0]))
            hsize = int((float(img.size[1]) * float(wpercent)))
            img = img.resize((basewidth, hsize), Image.Resampling.LANCZOS)
        else:
            img = img.resize((basewidth, baseheight), Image.Resampling.LANCZOS)

        if img.mode != 'P':
            img = img.convert('RGB')
        return img

    def add_corners(img, rad):
        circle = Image.new('L', (rad * 2, rad * 2), 0)
        draw = ImageDraw.Draw(circle)
        draw.ellipse((0, 0, rad * 2, rad * 2), fill=255)
        alpha = Image.new('L', img.size, "white")
        w, h = img.size
        alpha.paste(circle.crop((0, 0, rad, rad)), (0, 0))
        alpha.paste(circle.crop((0, rad, rad, rad * 2)), (0, h - rad))
        alpha.paste(circle.crop((rad, 0, rad * 2, rad)), (w - rad, 0))
        alpha.paste(circle.crop((rad, rad, rad * 2, rad * 2)), (w - rad, h - rad))
        img.putalpha(alpha)
        return img

    def load_image(url):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")  # suppress "corrupt exif" output from PIL
            try:
                img = Image.open(requests.get(url, stream=True).raw)
                if img.format == 'PNG' and img.mode != 'RGBA':
                    img = img.convert("RGBA")
                return img
            except Exception:
                return None

    def graphics_widget(img, image_format="block", corner_radius=0, colors=16777216) -> urwid.Widget:
        if not img:
            return urwid.SolidFill(fill_char=" ")

        if can_render_pixels(image_format) and corner_radius > 0:
            render_img = add_corners(img, 10)
        else:
            render_img = img

        return UrwidImage(get_base_image(render_img, image_format, colors), '<', upscale=True)
        # "<" means left-justify the image

except ImportError:
    from urwid.raw_display import Screen
    TuiScreen = Screen

    def image_support_enabled():
        return False

    def can_render_pixels(image_format: str):
        return False

    def get_base_image(image, image_format: str):
        return None

    def add_corners(img, rad):
        return None

    def load_image(url):
        return None

    def graphics_widget(img, image_format="block", corner_radius=0) -> urwid.Widget:
        return urwid.SolidFill(fill_char=" ")



import logging
import math
import urwid
import webbrowser

from typing import List, Optional

from toot.tui import app

from toot.tui.richtext import html_to_widgets, url_to_widget
from toot.utils.datetime import parse_datetime, time_ago
from toot.utils.language import language_name

from toot.entities import Status
from toot.tui.scroll import Scrollable, ScrollBar

from toot.tui.utils import highlight_keys
from toot.tui.images import image_support_enabled, graphics_widget, can_render_pixels
from toot.tui.widgets import SelectableText, SelectableColumns, RoundedLineBox


logger = logging.getLogger("toot")
screen = urwid.raw_display.Screen()


class Timeline(urwid.Columns):
    """
    Displays a list of statuses to the left, and status details on the right.
    """

    signals = [
        "close",  # Close thread
        "focus",  # Focus changed
        "next",   # Fetch more statuses
        "save",   # Save current timeline
    ]

    def __init__(
        self,
        tui: "app.TUI",
        name: str,
        statuses: List[Status],
        focus: int = 0,
        is_thread: bool = False
    ):
        self.tui = tui
        self.name = name
        self.is_thread = is_thread
        self.statuses = statuses
        self.status_list = self.build_status_list(statuses, focus=focus)
        self.can_render_pixels = can_render_pixels(self.tui.options.image_format)

        try:
            focused_status = statuses[focus]
        except IndexError:
            focused_status = None

        self.status_details = StatusDetails(self, focused_status)
        status_widget = self.wrap_status_details(self.status_details)

        super().__init__([
            ("weight", 40, self.status_list),
            ("weight", 0, urwid.AttrWrap(urwid.SolidFill("│"), "columns_divider")),
            ("weight", 60, status_widget),
        ])

    def wrap_status_details(self, status_details: "StatusDetails") -> urwid.Widget:
        """Wrap StatusDetails widget with a scrollbar and footer."""
        self.status_detail_scrollable = Scrollable(urwid.Padding(status_details, right=1))
        return urwid.Padding(
            urwid.Frame(
                body=ScrollBar(
                    self.status_detail_scrollable,
                    thumb_char="\u2588",
                    trough_char="\u2591",
                ),
                footer=self.get_option_text(status_details.status),
            ),
            left=1
        )

    def build_status_list(self, statuses, focus):
        items = [self.build_list_item(status) for status in statuses]
        walker = urwid.SimpleFocusListWalker(items)
        walker.set_focus(focus)
        urwid.connect_signal(walker, "modified", self.modified)
        return urwid.ListBox(walker)

    def build_list_item(self, status):
        item = StatusListItem(status, self.tui.options)
        urwid.connect_signal(item, "click", lambda *args:
            self.tui.show_context_menu(status))
        return urwid.AttrMap(item, None, focus_map={
            "status_list_account": "status_list_selected",
            "status_list_timestamp": "status_list_selected",
            "highlight": "status_list_selected",
            "dim": "status_list_selected",
            None: "status_list_selected",
        })

    def get_option_text(self, status: Optional[Status]) -> Optional[urwid.Text]:
        if not status:
            return None

        poll = status.original.data.get("poll")
        show_media = status.original.data["media_attachments"] and self.tui.options.media_viewer

        options = [
            "[A]ccount" if not status.is_mine else "",
            "[B]oost",
            "[D]elete" if status.is_mine else "",
            "[E]dit" if status.is_mine else "",
            "B[o]okmark",
            "[F]avourite",
            "[V]iew",
            "[T]hread" if not self.is_thread else "",
            "L[i]nks",
            "[M]edia" if show_media else "",
            "[R]eply",
            "[P]oll" if poll and not poll["expired"] else "",
            "So[u]rce",
            "[Z]oom",
            "Tra[n]slate" if self.tui.can_translate else "",
            "Cop[y]",
            "Help([?])",
        ]
        options = "\n" + " ".join(o for o in options if o)
        options = highlight_keys(options, "shortcut_highlight", "shortcut")
        return urwid.Text(options)

    def get_focused_status(self):
        try:
            return self.statuses[self.status_list.body.focus]
        except TypeError:
            return None

    def get_focused_status_with_counts(self):
        """Returns a tuple of:
            * focused status
            * focused status' index in the status list
            * length of the status list
        """
        return (
            self.get_focused_status(),
            self.status_list.body.focus,
            len(self.statuses),
        )

    def modified(self):
        """Called when the list focus switches to a new status"""
        status, index, count = self.get_focused_status_with_counts()

        if image_support_enabled:
            clear_op = getattr(self.tui.screen, "clear_images", None)
            # term-image's screen implementation has clear_images(),
            # urwid's implementation does not.
            # TODO: it would be nice not to check this each time thru

            if callable(clear_op):
                self.tui.screen.clear_images()

        self.draw_status_details(status)
        self._emit("focus")

    def refresh_status_details(self):
        """Redraws the details of the focused status."""
        status = self.get_focused_status()
        pos = self.status_detail_scrollable.get_scrollpos()
        self.draw_status_details(status)
        self.status_detail_scrollable.set_scrollpos(pos)

    def draw_status_details(self, status):
        self.status_details = StatusDetails(self, status)
        widget = self.wrap_status_details(self.status_details)
        self.contents[2] = widget, ("weight", 60, False)

    def keypress(self, size, key):
        status = self.get_focused_status()
        command = self._command_map[key]

        if not status:
            return super().keypress(size, key)

        # If down is pressed on last status in list emit a signal to load more.
        # TODO: Consider pre-loading statuses earlier
        if command in [urwid.CURSOR_DOWN, urwid.CURSOR_PAGE_DOWN] \
                and self.status_list.body.focus:
            index = self.status_list.body.focus + 1
            count = len(self.statuses)
            if index >= count:
                self._emit("next")

        if key in ("a", "A"):
            account_id = status.original.data["account"]["id"]
            self.tui.show_account(account_id)
            return

        if key in ("b", "B"):
            self.tui.async_toggle_reblog(self, status)
            return

        if key in ("c", "C"):
            self.tui.show_compose()
            return

        if key in ("d", "D"):
            if status.is_mine:
                self.tui.show_delete_confirmation(status)
            return

        if key in ("e", "E"):
            if status.is_mine:
                self.tui.async_edit(status)
            return

        if key in ("f", "F"):
            self.tui.async_toggle_favourite(self, status)
            return

        if key in ("m", "M"):
            self.tui.show_media(status)
            return

        if key in ("q", "Q"):
            self._emit("close")
            return

        if key == "esc" and self.is_thread:
            self._emit("close")
            return

        if key in ("r", "R"):
            self.tui.show_compose(status)
            return

        if key in ("s", "S"):
            status.original.show_sensitive = True
            self.refresh_status_details()
            return

        if key in ("o", "O"):
            self.tui.async_toggle_bookmark(self, status)
            return

        if key in ("i", "I"):
            self.tui.show_links(status)
            return

        if key in ("n", "N"):
            if self.tui.can_translate:
                self.tui.async_translate(self, status)
            return

        if key in ("t", "T"):
            self.tui.show_thread(status)
            return

        if key in ("u", "U"):
            self.tui.show_status_source(status)
            return

        if key in ("v", "V"):
            if status.original.url:
                webbrowser.open(status.original.url)
                # force a screen refresh; necessary with console browsers
                self.tui.clear_screen()
            return

        if key in ("e", "E"):
            self._emit("save", status)
            return

        if key in ("z", "Z"):
            self.tui.show_status_zoom(self.status_details)
            return

        if key in ("p", "P"):
            poll = status.original.data.get("poll")
            if poll and not poll["expired"]:
                self.tui.show_poll(status)
            return

        if key in ("y", "Y"):
            self.tui.copy_status(status)
            return

        return super().keypress(size, key)

    def append_status(self, status):
        self.statuses.append(status)
        self.status_list.body.append(self.build_list_item(status))

    def prepend_status(self, status):
        self.statuses.insert(0, status)
        self.status_list.body.insert(0, self.build_list_item(status))

    def append_statuses(self, statuses):
        for status in statuses:
            self.append_status(status)

    def get_status_index(self, id):
        # TODO: This is suboptimal, consider a better way
        for n, status in enumerate(self.statuses.copy()):
            if status.id == id:
                return n
        raise ValueError("Status with ID {} not found".format(id))

    def focus_status(self, status):
        index = self.get_status_index(status.id)
        self.status_list.body.set_focus(index)

    def update_status(self, status):
        """Overwrite status in list with the new instance and redraw."""
        index = self.get_status_index(status.id)
        assert self.statuses[index].id == status.id  # Sanity check

        # Update internal status list
        self.statuses[index] = status

        # Redraw list item
        self.status_list.body[index] = self.build_list_item(status)

        # Redraw status details if status is focused
        if index == self.status_list.body.focus:
            self.draw_status_details(status)

    def update_status_image(self, status, path, placeholder_index):
        """Replace image placeholder with image widget and redraw"""
        index = self.get_status_index(status.id)
        assert self.statuses[index].id == status.id  # Sanity check

        # get the image and replace the placeholder with a graphics widget
        img = None
        if hasattr(self, "images"):
            try:
                img = self.images[(str(hash(path)))]
            except KeyError:
                pass
        if img:
            try:
                status.placeholders[placeholder_index].original_widget = graphics_widget(
                    img, image_format=self.tui.options.image_format, corner_radius=10, colors=self.tui.options.colors
                )

            except IndexError:
                # ignore IndexErrors.
                pass

    def remove_status(self, status):
        index = self.get_status_index(status.id)
        assert self.statuses[index].id == status.id  # Sanity check

        del self.statuses[index]
        del self.status_list.body[index]
        self.refresh_status_details()


class StatusDetails(urwid.Pile):
    def __init__(self, timeline: Timeline, status: Optional[Status]):
        self.status = status
        self.timeline = timeline
        if self.status:
            self.status.placeholders = []
        self.followed_accounts = timeline.tui.followed_accounts
        self.options = timeline.tui.options

        reblogged_by = status.author if status and status.reblog else None
        widget_list = list(self.content_generator(status.original, reblogged_by)
            if status else ())
        return super().__init__(widget_list)

    def image_widget(self, path, rows=None, aspect=None) -> urwid.Widget:
        """Returns a widget capable of displaying the image

        path is required; URL to image
        rows, if specfied, sets a fixed number of rows. Or:
        aspect, if specified, calculates rows based on pane width
        and the aspect ratio provided"""

        if not rows:
            if not aspect:
                aspect = 3 / 2  # reasonable default

            screen_rows = screen.get_cols_rows()[1]
            if self.timeline.can_render_pixels:
                # for pixel-rendered images,
                # image rows should be 33% of the available screen
                # but in no case fewer than 10
                rows = max(10, math.floor(screen_rows * .33))
            else:
                # for cell-rendered images,
                # use the max available columns
                # and calculate rows based on the image
                # aspect ratio
                cols = math.floor(0.55 * screen.get_cols_rows()[0])
                rows = math.ceil((cols / 2) / aspect)
                # if the calculated rows are more than will
                # fit on one screen, reduce to one screen of rows
                rows = min(screen_rows - 6, rows)

                # but in no case fewer than 10 rows
                rows = max(rows, 10)

        img = None
        if hasattr(self.timeline, "images"):
            try:
                img = self.timeline.images[(str(hash(path)))]
            except KeyError:
                pass
        if img:
            return (urwid.BoxAdapter(
                graphics_widget(img, image_format=self.timeline.tui.options.image_format, corner_radius=10,
                                colors=self.timeline.tui.options.colors), rows))
        else:
            placeholder = urwid.BoxAdapter(urwid.SolidFill(fill_char=" "), rows)
            self.status.placeholders.append(placeholder)
            if image_support_enabled():
                self.timeline.tui.async_load_image(self.timeline, self.status, path, len(self.status.placeholders) - 1)
            return placeholder

    def author_header(self, reblogged_by):
        avatar_url = self.status.original.data["account"]["avatar"]

        if avatar_url and image_support_enabled():
            aimg = self.image_widget(avatar_url, 2)

        account_color = ("highlight" if self.status.original.author.account in
                        self.timeline.tui.followed_accounts else "account")

        atxt = urwid.Pile([("pack", urwid.Text(("bold", self.status.original.author.display_name))),
                           ("pack", urwid.Text((account_color, self.status.original.author.account)))])

        if image_support_enabled():
            columns = urwid.Columns([aimg, ("weight", 9999, atxt)], dividechars=1, min_width=5)
        else:
            columns = urwid.Columns([("weight", 9999, atxt)], dividechars=1, min_width=5)

        return columns

    def content_generator(self, status, reblogged_by):
        if reblogged_by:
            reblogger_name = (reblogged_by.display_name
                              if reblogged_by.display_name
                              else reblogged_by.username)
            text = f"♺ {reblogger_name} boosted"
            yield urwid.Text(("dim", text))
            yield ("pack", urwid.AttrMap(urwid.Divider("-"), "dim"))

        yield self.author_header(reblogged_by)
        yield ("pack", urwid.Divider())

        if status.data["spoiler_text"]:
            yield ("pack", urwid.Text(status.data["spoiler_text"]))
            yield ("pack", urwid.Divider())

        # Show content warning
        if status.data["spoiler_text"] and not status.show_sensitive and not self.options.always_show_sensitive:
            yield ("pack", urwid.Text(("content_warning", "Marked as sensitive. Press S to view.")))
        else:
            if status.data["spoiler_text"]:
                yield ("pack", urwid.Text(("content_warning", "Marked as sensitive.")))

            content = status.original.translation if status.original.show_translation else status.data["content"]
            widgetlist = html_to_widgets(content)

            for line in widgetlist:
                yield (line)

            media = status.data["media_attachments"]
            if media:
                for m in media:
                    yield ("pack", urwid.AttrMap(urwid.Divider("-"), "dim"))
                    yield ("pack", urwid.Text([("bold", "Media attachment"), " (", m["type"], ")"]))
                    if m["description"]:
                        yield ("pack", urwid.Text(m["description"]))
                    if m["url"]:
                        if m["url"].lower().endswith(('.jpg', '.jpeg', '.png', '.gif', '.svg', '.webp')):
                            yield urwid.Text("")
                            try:
                                aspect = float(m["meta"]["original"]["aspect"])
                            except Exception:
                                aspect = None
                            if image_support_enabled():
                                yield self.image_widget(m["url"], aspect=aspect)
                            yield urwid.Divider()
                        # video media may include a preview URL, show that as a fallback
                        elif m["preview_url"]:
                            if m["preview_url"].lower().endswith(('.jpg', '.jpeg', '.png', '.gif', '.svg', '.webp')):
                                yield urwid.Text("")
                                try:
                                    aspect = float(m["meta"]["small"]["aspect"])
                                except Exception:
                                    aspect = None
                                if image_support_enabled():
                                    yield self.image_widget(m["preview_url"], aspect=aspect)
                                yield urwid.Divider()
                        yield ("pack", url_to_widget(m["url"]))

            poll = status.original.data.get("poll")
            if poll:
                yield ("pack", urwid.Divider())
                yield ("pack", self.build_linebox(self.poll_generator(poll)))

            card = status.data.get("card")
            if card:
                yield ("pack", urwid.Divider())
                yield ("pack", self.build_linebox(self.card_generator(card)))

        application = status.data.get("application") or {}
        application = application.get("name")

        yield ("pack", urwid.AttrWrap(urwid.Divider("-"), "dim"))

        translated_from = (
            language_name(status.original.translated_from)
            if status.original.show_translation and status.original.translated_from
            else None
        )

        visibility_colors = {
            "public": "visibility_public",
            "unlisted": "visibility_unlisted",
            "private": "visibility_private",
            "direct": "visibility_direct"
        }

        visibility = status.visibility.title()
        visibility_color = visibility_colors.get(status.visibility, "dim")

        yield ("pack", urwid.Text([
            ("status_detail_timestamp", f"{status.created_at.strftime('%Y-%m-%d %H:%M')} "),
            ("status_detail_timestamp",
             f"(edited {status.edited_at.strftime('%Y-%m-%d %H:%M')}) " if status.edited_at else ""),
            ("status_detail_bookmarked" if status.bookmarked else "dim", "b "),
            ("dim", f"⤶ {status.data['replies_count']} "),
            ("highlight" if status.reblogged else "dim", f"♺ {status.data['reblogs_count']} "),
            ("highlight" if status.favourited else "dim", f"★ {status.data['favourites_count']}"),
            (visibility_color, f" · {visibility}"),
            ("highlight", f" · Translated from {translated_from} " if translated_from else ""),
            ("dim", f" · {application}" if application else ""),
        ]))

        # Push things to bottom
        yield ("weight", 1, urwid.BoxAdapter(urwid.SolidFill(" "), 1))

    def build_linebox(self, contents):
        contents = urwid.Pile(list(contents))
        contents = urwid.Padding(contents, left=1, right=1)
        return RoundedLineBox(contents)

    def card_generator(self, card):
        yield urwid.Text(("card_title", card["title"].strip()))
        if card.get("author_name"):
            yield urwid.Text(["by ", ("card_author", card["author_name"].strip())])
        yield urwid.Text("")
        if card["description"]:
            yield urwid.Text(card["description"].strip())
            yield urwid.Text("")
        yield url_to_widget(card["url"])

        if card["image"] and image_support_enabled():
            if card["image"].lower().endswith(('.jpg', '.jpeg', '.png', '.gif', '.svg', '.webp')):
                yield urwid.Text("")
                try:
                    aspect = int(card["width"]) / int(card["height"])
                except Exception:
                    aspect = None
                yield self.image_widget(card["image"], aspect=aspect)

    def poll_generator(self, poll):
        for idx, option in enumerate(poll["options"]):
            perc = (round(100 * option["votes_count"] / poll["votes_count"])
                if poll["votes_count"] else 0)

            if poll["voted"] and poll["own_votes"] and idx in poll["own_votes"]:
                voted_for = " ✓"
            else:
                voted_for = ""

            yield urwid.Text(option["title"] + voted_for)
            yield urwid.ProgressBar("", "poll_bar", perc)

        status = "Poll · {} votes".format(poll["votes_count"])

        if poll["expired"]:
            status += " · Closed"

        if poll["expires_at"]:
            expires_at = parse_datetime(poll["expires_at"]).strftime("%Y-%m-%d %H:%M")
            status += " · Closes on {}".format(expires_at)

        yield urwid.Text(("dim", status))


class StatusListItem(SelectableColumns):
    def __init__(self, status, options):
        edited_at = status.original.edited_at

        # TODO: hacky implementation to avoid creating conflicts for existing
        # pull requests, refactor when merged.
        created_at = (
            time_ago(status.created_at).ljust(3, " ")
            if options.relative_datetimes
            else status.created_at.strftime("%Y-%m-%d %H:%M")
        )

        edited_flag = "*" if edited_at else " "
        favourited = ("highlight", "★") if status.original.favourited else " "
        reblogged = ("highlight", "♺") if status.original.reblogged else " "
        is_reblog = ("dim", "♺") if status.reblog else " "
        is_reply = ("dim", "⤶ ") if status.original.in_reply_to else "  "

        display_name = status.original.author.display_name
        account_name = status.original.account
        name = display_name if display_name and options.show_display_names else account_name

        return super().__init__([
            ("pack", SelectableText(("status_list_timestamp", created_at), wrap="clip")),
            ("pack", urwid.Text(("status_list_timestamp", edited_flag))),
            ("pack", urwid.Text(" ")),
            ("pack", urwid.Text(favourited)),
            ("pack", urwid.Text(" ")),
            ("pack", urwid.Text(reblogged)),
            ("pack", urwid.Text(" ")),
            urwid.Text(("status_list_account", name), wrap="clip"),
            ("pack", urwid.Text(is_reply)),
            ("pack", urwid.Text(is_reblog)),
            ("pack", urwid.Text(" ")),
        ])



import os
import sys

from os.path import join, expanduser
from typing import NamedTuple
from importlib import metadata


try:
    __version__ = metadata.version("toot")
except metadata.PackageNotFoundError:
    __version__ = "0.0.0"


class App(NamedTuple):
    instance: str
    base_url: str
    client_id: str
    client_secret: str


class User(NamedTuple):
    instance: str
    username: str
    access_token: str


DEFAULT_INSTANCE = 'https://mastodon.social'

CLIENT_NAME = 'toot - a Mastodon CLI client'
CLIENT_WEBSITE = 'https://github.com/ihabunek/toot'

TOOT_CONFIG_DIR_NAME = "toot"


def get_config_dir():
    """Returns the path to toot config directory"""

    # On Windows, store the config in roaming appdata
    if sys.platform == "win32" and "APPDATA" in os.environ:
        return join(os.getenv("APPDATA"), TOOT_CONFIG_DIR_NAME)

    # Respect XDG_CONFIG_HOME env variable if set
    # https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
    if "XDG_CONFIG_HOME" in os.environ:
        config_home = expanduser(os.environ["XDG_CONFIG_HOME"])
        return join(config_home, TOOT_CONFIG_DIR_NAME)

    # Default to ~/.config/toot/
    return join(expanduser("~"), ".config", TOOT_CONFIG_DIR_NAME)



import platform
import click
import os
import re
import subprocess
import tempfile
import unicodedata
import warnings

from bs4 import BeautifulSoup
from importlib.metadata import version
from itertools import islice
from typing import Any, Dict, Generator, Iterable, List, Optional, TypeVar
from urllib.parse import urlparse, urlencode, quote, unquote


def str_bool(b: bool) -> str:
    """Convert boolean to string, in the way expected by the API."""
    return "true" if b else "false"


def str_bool_nullable(b: Optional[bool]) -> Optional[str]:
    """Similar to str_bool, but leave None as None"""
    return None if b is None else str_bool(b)


def parse_html(html: str) -> BeautifulSoup:
    # Ignore warnings made by BeautifulSoup, if passed something that looks like
    # a file (e.g. a dot which matches current dict), it will warn that the file
    # should be opened instead of passing a filename.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return BeautifulSoup(html.replace("&apos;", "'"), "html.parser")


def get_text(html: str) -> str:
    """Converts html to text, strips all tags."""
    text = parse_html(html).get_text()
    return unicodedata.normalize("NFKC", text)


def html_to_paragraphs(html: str) -> List[List[str]]:
    """Attempt to convert html to plain text while keeping line breaks.
    Returns a list of paragraphs, each being a list of lines.
    """
    paragraphs = re.split("</?p[^>]*>", html)

    # Convert <br>s to line breaks and remove empty paragraphs
    paragraphs = [re.split("<br */?>", p) for p in paragraphs if p]

    # Convert each line in each paragraph to plain text:
    return [[get_text(line) for line in p] for p in paragraphs]


def format_content(content: str) -> Generator[str, None, None]:
    """Given a Status contents in HTML, converts it into lines of plain text.

    Returns a generator yielding lines of content.
    """

    paragraphs = html_to_paragraphs(content)

    first = True

    for paragraph in paragraphs:
        if not first:
            yield ""

        for line in paragraph:
            yield line

        first = False


EOF_KEY = "Ctrl-Z" if os.name == 'nt' else "Ctrl-D"


def multiline_input() -> str:
    """Lets user input multiple lines of text, terminated by EOF."""
    lines: List[str] = []
    while True:
        try:
            lines.append(input())
        except EOFError:
            break

    return "\n".join(lines).strip()


EDITOR_DIVIDER = "------------------------ >8 ------------------------"

EDITOR_INPUT_INSTRUCTIONS = f"""
{EDITOR_DIVIDER}
Do not modify or remove the line above.
Enter your toot above it.
Everything below it will be ignored.
"""


def editor_input(editor: str, initial_text: str) -> str:
    """Lets user input text using an editor."""
    tmp_path = _tmp_status_path()
    initial_text = (initial_text or "") + EDITOR_INPUT_INSTRUCTIONS

    if not _use_existing_tmp_file(tmp_path):
        with open(tmp_path, "w") as f:
            f.write(initial_text)
            f.flush()

    subprocess.run([editor, tmp_path])

    with open(tmp_path) as f:
        return f.read().split(EDITOR_DIVIDER)[0].strip()


def delete_tmp_status_file() -> None:
    try:
        os.unlink(_tmp_status_path())
    except FileNotFoundError:
        pass


def _tmp_status_path() -> str:
    tmp_dir = tempfile.gettempdir()
    return f"{tmp_dir}/.status.toot"


def _use_existing_tmp_file(tmp_path: str) -> bool:
    if os.path.exists(tmp_path):
        click.echo(f"Found draft status at: {tmp_path}")

        choice = click.Choice(["O", "D"], case_sensitive=False)
        char = click.prompt("Open or Delete?", type=choice, default="O")
        return char == "O"

    return False


def drop_empty_values(data: Dict[Any, Any]) -> Dict[Any, Any]:
    """Remove keys whose values are null"""
    return {k: v for k, v in data.items() if v is not None}


def urlencode_url(url: str) -> str:
    parsed_url = urlparse(url)

    # unencode before encoding, to prevent double-urlencoding
    encoded_path = quote(unquote(parsed_url.path), safe="-._~()'!*:@,;+&=/")
    encoded_query = urlencode({k: quote(unquote(v), safe="-._~()'!*:@,;?/") for k, v in parsed_url.params})
    encoded_url = parsed_url._replace(path=encoded_path, params=encoded_query).geturl()

    return encoded_url


def get_distro_name() -> Optional[str]:
    """Attempt to get linux distro name from platform (requires python 3.10+)"""
    try:
        return platform.freedesktop_os_release()["PRETTY_NAME"]  # type: ignore  # novermin
    except Exception:
        pass


def get_version(name):
    try:
        return version(name)
    except Exception:
        return None


T = TypeVar("T")


def batched(iterable: Iterable[T], n: int) -> Generator[List[T], None, None]:
    """Batch data from the iterable into lists of length n. The last batch may
    be shorter than n."""
    if n < 1:
        raise ValueError("n must be positive")
    iterator = iter(iterable)
    while True:
        batch = list(islice(iterator, n))
        if batch:
            yield batch
        else:
            break



import logging
import subprocess
import urwid


from concurrent.futures import ThreadPoolExecutor
from typing import NamedTuple, Optional
from datetime import datetime, timezone

from toot import api, config, __version__, settings
from toot import App, User
from toot.cli import get_default_visibility
from toot.utils.datetime import parse_datetime

from .compose import StatusComposer
from .constants import PALETTE
from .entities import Status
from .images import TuiScreen, load_image
from .overlays import ExceptionStackTrace, GotoMenu, Help, StatusSource, StatusLinks, StatusZoom
from .overlays import StatusDeleteConfirmation, Account
from .poll import Poll
from .timeline import Timeline
from .utils import get_max_toot_chars, parse_content_links, copy_to_clipboard, LRUCache
from .widgets import ModalBox, RoundedLineBox

logger = logging.getLogger(__name__)

urwid.set_encoding('UTF-8')


DEFAULT_MAX_TOOT_CHARS = 500


class TuiOptions(NamedTuple):
    colors: int
    media_viewer: Optional[str]
    always_show_sensitive: bool
    relative_datetimes: bool
    cache_size: int
    default_visibility: Optional[str]
    image_format: Optional[str]
    show_display_names: bool


class Header(urwid.WidgetWrap):
    def __init__(self, app, user):
        self.app = app
        self.user = user

        self.text = urwid.Text("")
        self.cols = urwid.Columns([
            ("pack", urwid.Text(('header_bold', 'toot'))),
            ("pack", urwid.Text(('header', ' | {}@{}'.format(user.username, app.instance)))),
            ("pack", self.text),
        ])

        widget = urwid.AttrMap(self.cols, 'header')
        widget = urwid.Padding(widget)
        self._wrapped_widget = widget

    def clear_text(self, text):
        self.text.set_text("")

    def set_text(self, text):
        self.text.set_text(" | " + text)


class Footer(urwid.Pile):
    def __init__(self):
        self.status = urwid.Text("")
        self.message = urwid.Text("")

        return super().__init__([
            urwid.AttrMap(self.status, "footer_status"),
            urwid.AttrMap(self.message, "footer_message"),
        ])

    def set_status(self, text):
        self.status.set_text(text)

    def clear_status(self, text):
        self.status.set_text("")

    def set_message(self, text):
        self.message.set_text(text)

    def set_error_message(self, text):
        self.message.set_text(("footer_message_error", text))

    def clear_message(self):
        self.message.set_text("")


class TUI(urwid.Frame):
    """Main TUI frame."""
    loop: urwid.MainLoop
    screen: urwid.BaseScreen

    @staticmethod
    def create(app: App, user: User, options: TuiOptions):
        """Factory method, sets up TUI and an event loop."""
        screen = TuiScreen()
        screen.set_terminal_properties(options.colors)

        tui = TUI(app, user, screen, options)

        palette = PALETTE.copy()
        overrides = settings.get_setting("tui.palette", dict, {})
        for name, styles in overrides.items():
            palette.append(tuple([name] + styles))

        loop = urwid.MainLoop(
            tui,
            palette=palette,
            event_loop=urwid.AsyncioEventLoop(),
            unhandled_input=tui.unhandled_input,
            screen=screen,
        )
        tui.loop = loop

        return tui

    def __init__(self, app, user, screen, options: TuiOptions):
        self.app = app
        self.user = user
        self.config = config.load_config()
        self.options = options

        self.loop = None  # late init, set in `create`
        self.screen = screen
        self.executor = ThreadPoolExecutor(max_workers=1)
        self.timeline_generator = api.home_timeline_generator(app, user, limit=40)

        # Show intro screen while toots are being loaded
        self.body = self.build_intro()
        self.header = Header(app, user)
        self.footer = Footer()
        self.footer.set_status("Loading...")

        # Default max status length, updated on startup
        self.max_toot_chars = DEFAULT_MAX_TOOT_CHARS

        self.timeline = None
        self.overlay = None
        self.exception = None
        self.can_translate = False
        self.account = None
        self.followed_accounts = []
        self.preferences = {}

        if self.options.cache_size:
            self.cache_max = 1024 * 1024 * self.options.cache_size
        else:
            self.cache_max = 1024 * 1024 * 10  # default 10MB

        super().__init__(self.body, header=self.header, footer=self.footer)

    def run(self):
        self.loop.set_alarm_in(0, lambda *args: self.async_load_account())
        self.loop.set_alarm_in(0, lambda *args: self.async_load_instance())
        self.loop.set_alarm_in(0, lambda *args: self.async_load_preferences())
        self.loop.set_alarm_in(0, lambda *args: self.async_load_timeline(
            is_initial=True, timeline_name="home"))
        self.loop.run()
        self.executor.shutdown(wait=False)

    def build_intro(self):
        font = urwid.font.Thin6x6Font()

        # NB: Padding with width="clip" will convert the fixed BigText widget
        # to a flow widget so it can be used in a Pile.

        big_text = "Toot {}".format(__version__)
        big_text = urwid.BigText(("intro_bigtext", big_text), font)
        big_text = urwid.Padding(big_text, align="center", width="clip")

        intro = urwid.Pile([
            big_text,
            urwid.Divider(),
            urwid.Text([
                "Maintained by ",
                ("intro_smalltext", "@ihabunek"),
                " and contributors"
            ], align="center"),
            urwid.Divider(),
            urwid.Text(("intro_smalltext", "Loading toots..."), align="center"),
        ])

        return urwid.Filler(intro)

    def run_in_thread(self, fn, done_callback=None, error_callback=None):
        """Runs `fn` asynchronously in a separate thread.

        On completion calls `done_callback` if `fn` exited cleanly, or
        `error_callback` if an exception was caught. Callback methods are
        invoked in the main thread, not the thread in which `fn` is executed.
        """

        def _default_error_callback(ex):
            self.exception = ex
            self.footer.set_error_message("An exception occurred, press X to view")

        _error_callback = error_callback or _default_error_callback

        def _done(future):
            try:
                result = future.result()
                if done_callback:
                    # Use alarm to invoke callback in main thread
                    self.loop.set_alarm_in(0, lambda *args: done_callback(result))
            except Exception as ex:
                exception = ex
                logger.exception(exception)
                self.loop.set_alarm_in(0, lambda *args: _error_callback(exception))

        # TODO: replace by `self.loop.event_loop.run_in_executor` at some point
        # Added in https://github.com/urwid/urwid/issues/575
        # Not yet released at the time of this comment
        future = self.loop.event_loop._loop.run_in_executor(self.executor, fn)
        future.add_done_callback(_done)
        return future

    def connect_default_timeline_signals(self, timeline):
        urwid.connect_signal(timeline, "focus", self.refresh_footer)

    def build_timeline(self, name, statuses, local):
        def _close(*args):
            raise urwid.ExitMainLoop()

        def _next(*args):
            self.async_load_timeline(is_initial=False)

        def _toggle_save(timeline, status):
            if not timeline.name.startswith("#"):
                return
            hashtag = timeline.name[1:]
            assert isinstance(local, bool), local
            timelines = self.config.setdefault("timelines", {})
            if hashtag in timelines:
                del timelines[hashtag]
                self.footer.set_message("#{} unpinned".format(hashtag))
            else:
                timelines[hashtag] = {"local": local}
                self.footer.set_message("#{} pinned".format(hashtag))
            self.loop.set_alarm_in(5, lambda *args: self.footer.clear_message())
            config.save_config(self.config)

        timeline = Timeline(self, name, statuses)

        self.connect_default_timeline_signals(timeline)
        urwid.connect_signal(timeline, "next", _next)
        urwid.connect_signal(timeline, "close", _close)
        urwid.connect_signal(timeline, "save", _toggle_save)

        return timeline

    def make_status(self, status_data):
        is_mine = self.user.username == status_data["account"]["acct"]
        return Status(status_data, is_mine, self.app.instance)

    def show_thread(self, status):
        def _close(*args):
            """When thread is closed, go back to the main timeline."""
            self.body = self.timeline
            self.body.refresh_status_details()
            self.refresh_footer(self.timeline)

        # This is pretty fast, so it's probably ok to block while context is
        # loaded, can be made async later if needed
        context = api.context(self.app, self.user, status.original.id).json()
        ancestors = [self.make_status(s) for s in context["ancestors"]]
        descendants = [self.make_status(s) for s in context["descendants"]]
        statuses = ancestors + [status] + descendants
        focus = len(ancestors)

        timeline = Timeline(self, "thread", statuses, focus=focus, is_thread=True)

        self.connect_default_timeline_signals(timeline)
        urwid.connect_signal(timeline, "close", _close)
        self.body = timeline
        timeline.refresh_status_details()
        self.refresh_footer(timeline)

    def async_load_timeline(self, is_initial, timeline_name=None, local=None):
        """Asynchronously load a list of statuses."""

        def _load_statuses():
            self.footer.set_message("Loading statuses...")
            try:
                data = next(self.timeline_generator)
            except StopIteration:
                return []
            finally:
                self.footer.clear_message()

            return [self.make_status(s) for s in data]

        def _done_initial(statuses):
            """Process initial batch of statuses, construct a Timeline."""
            self.timeline = self.build_timeline(timeline_name, statuses, local)
            self.timeline.refresh_status_details()  # Draw first status
            self.refresh_footer(self.timeline)
            self.body = self.timeline

        def _done_next(statuses):
            """Process sequential batch of statuses, adds statuses to the
            existing timeline."""
            self.timeline.append_statuses(statuses)

        return self.run_in_thread(_load_statuses,
            done_callback=_done_initial if is_initial else _done_next)

    def async_load_instance(self):
        """
        Attempt to update max_toot_chars from instance data.
        Does not work on vanilla Mastodon, works on Pleroma.
        See: https://github.com/tootsuite/mastodon/issues/4915

        Also attempt to update translation flag from instance
        data. Translation is only present on Mastodon 4+ servers
        where the administrator has enabled this feature.
        See: https://github.com/mastodon/mastodon/issues/19328
        """
        def _load_instance():
            return api.get_instance(self.app.base_url).json()

        def _done(instance):
            self.max_toot_chars = get_max_toot_chars(instance, DEFAULT_MAX_TOOT_CHARS)
            logger.info(f"Max toot chars set to: {self.max_toot_chars}")

            if "translation" in instance:
                # instance is advertising translation service
                self.can_translate = instance["translation"]["enabled"]
            elif "version" in instance:
                # fallback check:
                # get the major version number of the server
                # this works for Mastodon and Pleroma version strings
                # Mastodon versions < 4 do not have translation service
                # If the version is missing, assume 0 as a fallback
                # Revisit this logic if Pleroma implements translation
                version = instance["version"]
                ch = "0" if not version else version[0]
                self.can_translate = int(ch) > 3 if ch.isnumeric() else False

        return self.run_in_thread(_load_instance, done_callback=_done)

    def async_load_preferences(self):
        """
        Attempt to update user preferences from instance.
        https://docs.joinmastodon.org/methods/preferences/
        """
        def _load_preferences():
            return api.get_preferences(self.app, self.user).json()

        def _done(preferences):
            self.preferences = preferences

        return self.run_in_thread(_load_preferences, done_callback=_done)

    def async_load_account(self):
        def _load_account():
            return api.verify_credentials(self.app, self.user).json()

        def _done_account(account):
            self.account = account
            self.loop.set_alarm_in(0, lambda *_: self.async_load_followed_accounts())

        self.run_in_thread(_load_account, done_callback=_done_account)

    def async_load_followed_accounts(self):
        def _load_accounts():
            try:
                if self.account:
                    return api.following(self.app, self.user, self.account["id"])
                return []
            except Exception:
                # not supported by all Mastodon servers so fail silently if necessary
                return []

        def _done_accounts(accounts):
            self.followed_accounts = {a["acct"] for a in accounts}

        self.run_in_thread(_load_accounts, done_callback=_done_accounts)

    def refresh_footer(self, timeline):
        """Show status details in footer."""
        status, index, count = timeline.get_focused_status_with_counts()
        self.footer.set_status([
            ("footer_status_bold", "[{}] ".format(timeline.name)),
        ] + ([status.id, " - status ", str(index + 1), " of ", str(count)]
            if status else ["no focused status"]))

    def show_status_source(self, status):
        self.open_overlay(
            widget=StatusSource(status),
            title="Status source",
        )

    def clear_screen(self):
        self.screen.clear()

    def show_links(self, status):
        links = parse_content_links(status.original.data["content"]) if status else []
        post_attachments = status.original.data["media_attachments"] or []
        reblog_attachments = (status.data["reblog"]["media_attachments"] if status.data["reblog"] else None) or []

        for a in post_attachments + reblog_attachments:
            url = a["remote_url"] or a["url"]
            links.append((url, a["description"] if a["description"] else url))

        def _clear(*args):
            self.clear_screen()

        if links:
            links = list(set(links))  # deduplicate links
            links = sorted(links, key=lambda link: link[0])  # sort alphabetically by URL
            sl_widget = StatusLinks(links)
            urwid.connect_signal(sl_widget, "clear-screen", _clear)
            self.open_overlay(
                widget=sl_widget,
                title="Status links",
                options={"height": len(links) + 2},
            )

    def show_status_zoom(self, status_details):
        self.open_overlay(
            widget=StatusZoom(status_details),
            title="Status zoom",
        )

    def show_exception(self, exception):
        self.open_overlay(
            widget=ExceptionStackTrace(exception),
            title="Unhandled Exception",
        )

    def show_compose(self, in_reply_to=None):
        def _close(*args):
            self.close_overlay()

        def _post(timeline, *args):
            self.post_status(*args)

        # If the user specified --default-visibility, use that; otherwise,
        # try to use the server-side default visibility.  If that fails, fall
        # back to get_default_visibility().
        visibility = (self.options.default_visibility or
                      self.preferences.get('posting:default:visibility',
                                           get_default_visibility()))

        composer = StatusComposer(self.max_toot_chars, self.user.username,
                                  visibility, in_reply_to)
        urwid.connect_signal(composer, "close", _close)
        urwid.connect_signal(composer, "post", _post)
        self.open_overlay(composer, title="Compose status")

    def async_edit(self, status):
        def _fetch_source():
            return api.fetch_status_source(self.app, self.user, status.id).json()

        def _done(source):
            self.close_overlay()
            self.show_edit(status, source)

        please_wait = ModalBox("Loading status...")
        self.open_overlay(please_wait)

        self.run_in_thread(_fetch_source, done_callback=_done)

    def show_edit(self, status, source):
        def _close(*args):
            self.close_overlay()

        def _edit(timeline, *args):
            self.edit_status(status, *args)

        composer = StatusComposer(self.max_toot_chars, self.user.username,
                                  visibility=None, edit=status, source=source)
        urwid.connect_signal(composer, "close", _close)
        urwid.connect_signal(composer, "post", _edit)
        self.open_overlay(composer, title="Edit status")

    def show_goto_menu(self):
        user_timelines = self.config.get("timelines", {})
        user_lists = api.get_lists(self.app, self.user) or []

        menu = GotoMenu(user_timelines, user_lists)
        urwid.connect_signal(menu, "home_timeline",
            lambda x: self.goto_home_timeline())
        urwid.connect_signal(menu, "public_timeline",
            lambda x, local: self.goto_public_timeline(local))
        urwid.connect_signal(menu, "bookmark_timeline",
            lambda x, local: self.goto_bookmarks())
        urwid.connect_signal(menu, "notification_timeline",
            lambda x, local: self.goto_notifications())
        urwid.connect_signal(menu, "conversation_timeline",
            lambda x, local: self.goto_conversations())
        urwid.connect_signal(menu, "personal_timeline",
            lambda x, local: self.goto_personal_timeline())
        urwid.connect_signal(menu, "hashtag_timeline",
            lambda x, tag, local: self.goto_tag_timeline(tag, local=local))
        urwid.connect_signal(menu, "list_timeline",
            lambda x, list_item: self.goto_list_timeline(list_item))

        self.open_overlay(menu, title="Go to", options=dict(
            align="center", width=("relative", 60),
            valign="middle", height=18 + len(user_timelines) + len(user_lists),
        ))

    def show_help(self):
        self.open_overlay(Help(), title="Help")

    def show_poll(self, status):
        self.open_overlay(
            widget=Poll(self.app, self.user, status),
            title="Poll",
        )

    def goto_home_timeline(self):
        self.timeline_generator = api.home_timeline_generator(
            self.app, self.user, limit=40)
        promise = self.async_load_timeline(is_initial=True, timeline_name="home")
        promise.add_done_callback(lambda *args: self.close_overlay())

    def goto_public_timeline(self, local):
        self.timeline_generator = api.public_timeline_generator(
            self.app, self.user, local=local, limit=40)
        timeline_name = "local public" if local else "global public"
        promise = self.async_load_timeline(is_initial=True, timeline_name=timeline_name)
        promise.add_done_callback(lambda *args: self.close_overlay())

    def goto_bookmarks(self):
        self.timeline_generator = api.bookmark_timeline_generator(
            self.app, self.user, limit=40)
        promise = self.async_load_timeline(is_initial=True, timeline_name="bookmarks")
        promise.add_done_callback(lambda *args: self.close_overlay())

    def goto_notifications(self):
        self.timeline_generator = api.notification_timeline_generator(
            self.app, self.user, limit=40)
        promise = self.async_load_timeline(is_initial=True, timeline_name="notifications")
        promise.add_done_callback(lambda *args: self.close_overlay())

    def goto_conversations(self):
        self.timeline_generator = api.conversation_timeline_generator(
            self.app, self.user, limit=40
        )
        promise = self.async_load_timeline(
            is_initial=True, timeline_name="conversations"
        )
        promise.add_done_callback(lambda *args: self.close_overlay())

    def goto_tag_timeline(self, tag, local):
        self.timeline_generator = api.tag_timeline_generator(
            self.app, self.user, tag, local=local, limit=40)
        promise = self.async_load_timeline(
            is_initial=True, timeline_name="#{}".format(tag), local=local,
        )
        promise.add_done_callback(lambda *args: self.close_overlay())

    def goto_personal_timeline(self):
        if not self.account:
            return
        self.timeline_generator = api.account_timeline_generator_by_id(
            self.app, self.user, self.account["id"], reblogs=True, limit=40)
        promise = self.async_load_timeline(
            is_initial=True, timeline_name=f"personal {self.user.username}")
        promise.add_done_callback(lambda *args: self.close_overlay())

    def goto_list_timeline(self, list_item):
        self.timeline_generator = api.timeline_list_generator(
            self.app, self.user, list_item['id'], limit=40)
        promise = self.async_load_timeline(
            is_initial=True, timeline_name=f"\N{clipboard}{list_item['title']}")
        promise.add_done_callback(lambda *args: self.close_overlay())

    def show_media(self, status):
        urls = [m["url"] for m in status.original.data["media_attachments"]]
        if not urls:
            return

        media_viewer = self.options.media_viewer
        if media_viewer:
            try:
                subprocess.run([media_viewer] + urls)
            except FileNotFoundError:
                self.footer.set_error_message(f"Media viewer not found: '{media_viewer}'")
            except Exception as ex:
                self.exception = ex
                self.footer.set_error_message("Failed invoking media viewer. Press X to see exception.")
        else:
            self.footer.set_error_message("Media viewer not configured")

    def show_context_menu(self, status):
        # TODO: show context menu
        pass

    def show_delete_confirmation(self, status):
        def _delete(widget):
            promise = self.async_delete_status(self.timeline, status)
            promise.add_done_callback(lambda *args: self.close_overlay())

        def _close(widget):
            self.close_overlay()

        widget = StatusDeleteConfirmation(status)
        urwid.connect_signal(widget, "close", _close)
        urwid.connect_signal(widget, "delete", _delete)
        self.open_overlay(widget, title="Delete status?", options=dict(
            align="center", width=30,
            valign="middle", height=4,
        ))

    def post_status(self, content, warning, visibility, in_reply_to_id):
        data = api.post_status(
            self.app,
            self.user,
            content,
            spoiler_text=warning,
            visibility=visibility,
            in_reply_to_id=in_reply_to_id
        ).json()

        status = self.make_status(data)

        # TODO: fetch new items from the timeline?

        self.footer.set_message("Status posted {} \\o/".format(status.id))
        self.close_overlay()

    def edit_status(self, status, content, warning, visibility, in_reply_to_id):
        # We don't support editing polls (yet), so to avoid losing the poll
        # data from the original toot, copy it to the edit request.
        poll_args = {}
        poll = status.original.data.get('poll', None)

        if poll is not None:
            poll_args['poll_options'] = [o['title'] for o in poll['options']]
            poll_args['poll_multiple'] = poll['multiple']

            # Convert absolute expiry time into seconds from now.
            expires_at = parse_datetime(poll['expires_at'])
            expires_in = int((expires_at - datetime.now(timezone.utc)).total_seconds())
            poll_args['poll_expires_in'] = expires_in

            if 'hide_totals' in poll:
                poll_args['poll_hide_totals'] = poll['hide_totals']

        data = api.edit_status(
            self.app,
            self.user,
            status.id,
            content,
            spoiler_text=warning,
            visibility=visibility,
            **poll_args
        ).json()

        new_status = self.make_status(data)

        self.footer.set_message("Status edited {} \\o/".format(status.id))
        self.close_overlay()

        if self.timeline is not None:
            self.timeline.update_status(new_status)

    def show_account(self, account_id):
        account = api.whois(self.app, self.user, account_id)
        relationship = api.get_relationship(self.app, self.user, account_id)
        self.open_overlay(
            widget=Account(self.app, self.user, account, relationship, self.options),
            title="Account",
        )

    def async_toggle_favourite(self, timeline, status):
        def _favourite():
            api.favourite(self.app, self.user, status.id)

        def _unfavourite():
            api.unfavourite(self.app, self.user, status.id)

        def _done(loop):
            # Create a new Status with flipped favourited flag
            new_data = status.data
            new_data["favourited"] = not status.favourited
            new_status = self.make_status(new_data)
            timeline.update_status(new_status)

        self.run_in_thread(
            _unfavourite if status.favourited else _favourite,
            done_callback=_done
        )

    def async_toggle_reblog(self, timeline, status):
        def _reblog():
            api.reblog(self.app, self.user, status.original.id, visibility=get_default_visibility())

        def _unreblog():
            api.unreblog(self.app, self.user, status.original.id)

        def _done(loop):
            # Create a new Status with flipped reblogged flag
            new_data = status.data
            new_status = self.make_status(new_data)
            new_status.original.reblogged = not status.original.reblogged
            timeline.update_status(new_status)

        # Check if status is rebloggable
        no_reblog_because_private = status.visibility == "private" and not status.is_mine
        no_reblog_because_direct = status.visibility == "direct"
        if no_reblog_because_private or no_reblog_because_direct:
            self.footer.set_error_message("You may not reblog this {} status".format(status.visibility))
            return

        self.run_in_thread(
            _unreblog if status.original.reblogged else _reblog,
            done_callback=_done
        )

    def async_translate(self, timeline, status):
        def _translate():
            self.footer.set_message("Translating status {}".format(status.original.id))

            try:
                response = api.translate(self.app, self.user, status.original.id)
                if response["content"]:
                    self.footer.set_message("Status translated")
                else:
                    self.footer.set_error_message("Server returned empty translation")
                    response = None
            except Exception:
                response = None
                self.footer.set_error_message("Translate server error")

            self.loop.set_alarm_in(3, lambda *args: self.footer.clear_message())
            return response

        def _done(response):
            if response is not None:
                status.original.translation = response["content"]
                status.original.translated_from = response["detected_source_language"]
                status.original.show_translation = True
                timeline.update_status(status)

        # If already translated, toggle showing translation
        if status.original.translation:
            status.original.show_translation = not status.original.show_translation
            timeline.update_status(status)
        else:
            self.run_in_thread(_translate, done_callback=_done)

    def async_toggle_bookmark(self, timeline, status):
        def _bookmark():
            api.bookmark(self.app, self.user, status.id)

        def _unbookmark():
            api.unbookmark(self.app, self.user, status.id)

        def _done(loop):
            # Create a new Status with flipped bookmarked flag
            new_data = status.data
            new_data["bookmarked"] = not status.bookmarked
            new_status = self.make_status(new_data)
            timeline.update_status(new_status)

        self.run_in_thread(
            _unbookmark if status.bookmarked else _bookmark,
            done_callback=_done
        )

    def async_delete_status(self, timeline, status):
        def _delete():
            api.delete_status(self.app, self.user, status.id)

        def _done(loop):
            timeline.remove_status(status)

        return self.run_in_thread(_delete, done_callback=_done)

    def async_load_image(self, timeline, status, path, placeholder_index):
        def _load():
            # don't bother loading images for statuses we are not viewing now
            if timeline.get_focused_status().id != status.id:
                return

            if not hasattr(timeline, "images"):
                timeline.images = LRUCache(cache_max_bytes=self.cache_max)

            img = load_image(path)
            if img:
                timeline.images[str(hash(path))] = img

        def _done(loop):
            # don't bother loading images for statuses we are not viewing now
            if timeline.get_focused_status().id != status.id:
                return
            timeline.update_status_image(status, path, placeholder_index)

        return self.run_in_thread(_load, done_callback=_done)

    def copy_status(self, status):
        # TODO: copy a better version of status content
        # including URLs
        copy_to_clipboard(self.screen, status.original.data["content"])
        self.footer.set_message(f"Status {status.original.id} copied")

    # --- Overlay handling -----------------------------------------------------

    default_overlay_options = dict(
        align="center", width=("relative", 80),
        valign="middle", height=("relative", 80),
    )

    def open_overlay(self, widget, options={}, title=""):
        top_widget = RoundedLineBox(widget, title=title)
        bottom_widget = self.body

        _options = self.default_overlay_options.copy()
        _options.update(options)

        self.overlay = urwid.Overlay(
            top_widget,
            bottom_widget,
            **_options
        )
        self.body = self.overlay

    def close_overlay(self):
        self.body = self.overlay.bottom_w
        self.overlay = None
        if self.timeline:
            self.timeline.refresh_status_details()

    def refresh_timeline(self):
        # No point in refreshing the bookmarks timeline
        # and we don't have a good way to refresh a
        # list timeline yet (no reference to list ID kept)
        if (not self.timeline
                or self.timeline.name == 'bookmarks'
                or self.timeline.name.startswith("\N{clipboard}")):
            return

        if self.timeline.name.startswith("#"):
            self.timeline_generator = api.tag_timeline_generator(
                self.app, self.user, self.timeline.name[1:], limit=40)
        elif self.timeline.name.startswith("\N{clipboard}"):
            self.timeline_generator = api.tag_timeline_generator(
                self.app, self.user, self.timeline.name[1:], limit=40)
        else:
            if self.timeline.name.endswith("public"):
                self.timeline_generator = api.public_timeline_generator(
                    self.app, self.user, local=self.timeline.name.startswith("local"), limit=40)
            elif self.timeline.name == "notifications":
                self.timeline_generator = api.notification_timeline_generator(
                    self.app, self.user, limit=40)
            elif self.timeline.name == "conversations":
                self.timeline_generator = api.conversation_timeline_generator(
                    self.app, self.user, limit=40)
            else:
                # default to home timeline
                self.timeline_generator = api.home_timeline_generator(
                    self.app, self.user, limit=40)

        self.async_load_timeline(is_initial=True, timeline_name=self.timeline.name)

    # --- Keys -----------------------------------------------------------------

    def unhandled_input(self, key):
        # TODO: this should not be in unhandled input
        if key in ('x', 'X'):
            if self.exception:
                self.show_exception(self.exception)

        elif key in ('g', 'G'):
            if not self.overlay:
                self.show_goto_menu()

        elif key == '?':
            if not self.overlay:
                self.show_help()

        elif key == ',':
            if not self.overlay:
                self.refresh_timeline()

        elif key == 'esc':
            if self.overlay:
                self.close_overlay()
            elif self.timeline.name != "home":
                # similar to goto_home_timeline() but without handling overlay (absent here)
                self.timeline_generator = api.home_timeline_generator(
                    self.app, self.user, limit=40)
                self.async_load_timeline(is_initial=True, timeline_name="home")

        elif key in ('q', 'Q'):
            if self.overlay:
                self.close_overlay()
            else:
                raise urwid.ExitMainLoop()



import click

from typing import Optional
from toot.cli import TUI_COLORS, VISIBILITY_CHOICES, IMAGE_FORMAT_CHOICES, Context, cli, pass_context
from toot.cli.validators import validate_tui_colors, validate_cache_size
from toot.tui.app import TUI, TuiOptions

COLOR_OPTIONS = ", ".join(TUI_COLORS.keys())


@cli.command()
@click.option(
    "-r", "--relative-datetimes",
    is_flag=True,
    help="Show relative datetimes in status list"
)
@click.option(
    "-m", "--media-viewer",
    help="Program to invoke with media URLs to display the media files, such as 'feh'"
)
@click.option(
    "-c", "--colors",
    callback=validate_tui_colors,
    help=f"""Number of colors to use, one of {COLOR_OPTIONS}, defaults to 16 if
             using --color, and 1 if using --no-color."""
)
@click.option(
    "-s", "--cache-size",
    callback=validate_cache_size,
    help="""Specify the image cache maximum size in megabytes. Default: 10MB.
      Minimum: 1MB."""
)
@click.option(
    "-v", "--default-visibility",
    type=click.Choice(VISIBILITY_CHOICES),
    help="Default visibility when posting new toots; overrides the server-side preference"
)
@click.option(
    "-s", "--always-show-sensitive",
    is_flag=True,
    help="Expand toots with content warnings automatically"
)
@click.option(
    "-f", "--image-format",
    type=click.Choice(IMAGE_FORMAT_CHOICES),
    help="Image output format; support varies across terminals. Default: block"
)
@click.option(
    "--show-display-names",
    is_flag=True,
    default=False,
    help="Show display names instead of account names in the list view."
)
@pass_context
def tui(
    ctx: Context,
    colors: Optional[int],
    media_viewer: Optional[str],
    always_show_sensitive: bool,
    relative_datetimes: bool,
    cache_size: Optional[int],
    default_visibility: Optional[str],
    image_format: Optional[str],
    show_display_names: bool,
):
    """Launches the toot terminal user interface"""
    if colors is None:
        colors = 16 if ctx.color else 1

    options = TuiOptions(
        colors=colors,
        media_viewer=media_viewer,
        relative_datetimes=relative_datetimes,
        cache_size=cache_size,
        default_visibility=default_visibility,
        always_show_sensitive=always_show_sensitive,
        image_format=image_format,
        show_display_names=show_display_names,
    )
    tui = TUI.create(ctx.app, ctx.user, options)
    tui.run()



from functools import wraps
from typing import Iterable, List, Optional, Tuple
from urllib.parse import quote

import click
from requests import Response

from toot import api, http
from toot.cli import Context, cli, json_option, pass_context
from toot.cli.lists import get_list_id
from toot.cli.validators import validate_instance
from toot.entities import (
    Account,
    Status,
    from_dict_list,
    from_response,
    from_response_list,
)
from toot.output import get_continue, get_max_width, get_terminal_height, print_timeline, status_lines
from toot.utils import drop_empty_values, str_bool_nullable


def common_timeline_options(func):
    @click.option(
        "--min-id",
        help="""All results returned will be lesser than this ID. In effect,
             sets an upper bound on results.""",
    )
    @click.option(
        "--max-id",
        help="""All results returned will be greater than this ID. In effect,
             sets a lower bound on results.""",
    )
    @click.option(
        "--since-id",
        help="""Returns results immediately newer than this ID. In effect,
             sets a cursor at this ID and paginates forward.""",
    )
    @click.option(
        "-l",
        "--limit",
        type=int,
        help="""Number of results to fetch per request. [default: 20] [max: 40].
             (limit and max may depend on your server)""",
    )
    @click.option(
        "-p/ ",
        "--pager/--no-pager",
        help="Page the results",
        default=True,
    )
    @click.option(
        "-c",
        "--clear/--no-clear",
        help="Clear the screen before printing. If paged, clear before each page.",
        default=True,
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


instance_option = click.option(
    "-i",
    "--instance",
    callback=validate_instance,
    help="""Domain or base URL of the instance, e.g. 'mastodon.social' or
         'https://mastodon.social'. If not given will display timeline of the
         logged in server.""",
)


@cli.group()
def timelines():
    """Show various timelines"""


@timelines.command()
@click.argument("account_name")
@common_timeline_options
@json_option
@pass_context
def account(
    ctx: Context,
    account_name: str,
    min_id: Optional[str],
    max_id: Optional[str],
    since_id: Optional[str],
    limit: Optional[int],
    pager: bool,
    clear: bool,
    json: bool,
):
    """View statuses posted to the given account."""
    response = api.lookup(ctx.app, ctx.user, account_name)
    account = from_response(Account, response)
    path = f"/api/v1/accounts/{account.id}/statuses"
    params = {
        "min_id": min_id,
        "max_id": max_id,
        "since_id": since_id,
        "limit": limit,
    }

    _show_timeline(ctx, path, params, json, pager, clear, limit)


@timelines.command()
@common_timeline_options
@json_option
@pass_context
def home(
    ctx: Context,
    min_id: Optional[str],
    max_id: Optional[str],
    since_id: Optional[str],
    limit: Optional[int],
    pager: bool,
    clear: bool,
    json: bool,
):
    """View statuses from followed users and hashtags."""
    path = "/api/v1/timelines/home"
    params = {
        "min_id": min_id,
        "max_id": max_id,
        "since_id": since_id,
        "limit": limit,
    }

    _show_timeline(ctx, path, params, json, pager, clear, limit)


@timelines.command()
@click.argument("link_url")
@common_timeline_options
@json_option
@pass_context
def link(
    ctx: Context,
    link_url: str,
    min_id: Optional[str],
    max_id: Optional[str],
    since_id: Optional[str],
    limit: Optional[int],
    pager: bool,
    clear: bool,
    json: bool,
):
    """View statuses sharing a link.

    View public statuses containing a link to the specified
    currently-trending article. This only lists statuses from people who have
    opted in to discoverability features."""
    path = "/api/v1/timelines/link"

    params = {
        "url": link_url,
        "min_id": min_id,
        "max_id": max_id,
        "since_id": since_id,
        "limit": limit,
    }

    _show_timeline(ctx, path, params, json, pager, clear, limit)


@timelines.command("list")
@click.argument("list_name_or_id")
@common_timeline_options
@json_option
@pass_context
def list_cmd(
    ctx: Context,
    list_name_or_id: str,
    min_id: Optional[str],
    max_id: Optional[str],
    since_id: Optional[str],
    limit: Optional[int],
    pager: bool,
    clear: bool,
    json: bool,
):
    """View statuses in the given list timeline."""
    list_id = get_list_id(ctx, list_name_or_id, list_name_or_id)
    path = f"/api/v1/timelines/list/{list_id}"

    params = {
        "min_id": min_id,
        "max_id": max_id,
        "since_id": since_id,
        "limit": limit,
    }

    _show_timeline(ctx, path, params, json, pager, clear, limit)


@timelines.command()
@common_timeline_options
@instance_option
@click.option(
    "--local",
    is_flag=True,
    default=None,
    help="Show only local statuses",
)
@click.option(
    "--remote",
    is_flag=True,
    default=None,
    help="Show only remote statuses",
)
@click.option(
    "--only-media",
    is_flag=True,
    default=None,
    help="Show only statuses with media attached",
)
@json_option
@pass_context
def public(
    ctx: Context,
    instance: Optional[str],
    min_id: Optional[str],
    max_id: Optional[str],
    since_id: Optional[str],
    limit: Optional[int],
    pager: bool,
    clear: bool,
    local: Optional[bool],
    remote: Optional[bool],
    only_media: Optional[bool],
    json: bool,
):
    """View public statuses."""
    path = "/api/v1/timelines/public"
    params = {
        "min_id": min_id,
        "max_id": max_id,
        "since_id": since_id,
        "limit": limit,
        "local": str_bool_nullable(local),
        "remote": str_bool_nullable(remote),
        "only_media": str_bool_nullable(only_media),
    }

    if instance:
        url = f"{instance}{path}"
        _show_anon_timeline(url, params, json, pager, clear, limit)
    else:
        _show_timeline(ctx, path, params, json, pager, clear, limit)


@timelines.command()
@common_timeline_options
@instance_option
@click.argument("tag_name")
@click.option(
    "--local",
    is_flag=True,
    default=None,
    help="Show only local statuses",
)
@click.option(
    "--remote",
    is_flag=True,
    default=None,
    help="Show only remote statuses",
)
@click.option(
    "--only-media",
    is_flag=True,
    default=None,
    help="Show only statuses with media attached",
)
@click.option(
    "--any",
    multiple=True,
    help="Return statuses that contain any of these additional tags (can be specified multiple times)",
)
@click.option(
    "--all",
    multiple=True,
    help="Return statuses that contain all of these additional tags (can be specified multiple times)",
)
@click.option(
    "--none",
    multiple=True,
    help="Return statuses that contain none of these additional tags (can be specified multiple times)",
)
@json_option
@pass_context
def tag(
    ctx: Context,
    tag_name: str,
    instance: Optional[str],
    min_id: Optional[str],
    max_id: Optional[str],
    since_id: Optional[str],
    limit: Optional[int],
    pager: bool,
    clear: bool,
    local: Optional[bool],
    remote: Optional[bool],
    only_media: Optional[bool],
    any: Tuple[str],
    all: Tuple[str],
    none: Tuple[str],
    json: bool,
):
    """View public statuses containing the given hashtag."""
    path = f"/api/v1/timelines/tag/{quote(tag_name)}"
    params = {
        "min_id": min_id,
        "max_id": max_id,
        "since_id": since_id,
        "limit": limit,
        "local": str_bool_nullable(local),
        "remote": str_bool_nullable(remote),
        "only_media": str_bool_nullable(only_media),
        "any[]": any or None,
        "all[]": all or None,
        "none[]": none or None,
    }

    if instance:
        url = f"{instance}{path}"
        _show_anon_timeline(url, params, json, pager, clear, limit)
    else:
        _show_timeline(ctx, path, params, json, pager, clear, limit)


def _show_timeline(ctx, path, params, json, pager, clear, limit):
    params = drop_empty_values(params)

    if json:
        response = http.get(ctx.app, ctx.user, path, params)
        click.echo(response.text)
        return

    if pager:
        responses = http.get_paged(ctx.app, ctx.user, path, params)
        _print_paged(responses, clear)
        return

    response = http.get(ctx.app, ctx.user, path, params)
    _print_single(response, clear, limit)


def _show_anon_timeline(url, params, json, pager, clear, limit):
    params = drop_empty_values(params)

    if json:
        response = http.anon_get(url, params)
        click.echo(response.text)
        return

    if pager:
        responses = http.anon_get_paged(url, params)
        _print_paged(responses, clear)
        return

    response = http.anon_get(url, params)
    _print_single(response, clear, limit)


def _print_single(response: Response, clear: bool, limit: Optional[int]):
    statuses = from_response_list(Status, response)
    if clear:
        click.clear()
    if statuses:
        print_timeline(statuses)
        if not limit or len(statuses) == limit:
            click.secho(
                "There may be more results. Increase the --limit or use --pager to see the rest.",
                dim=True,
            )
    else:
        click.echo("No statuses found")


def _print_paged(responses: Iterable[Response], clear: bool):
    width = get_max_width()
    height = get_terminal_height()
    separator = "─" * width

    def _page_generator():
        batch_lines: List[str] = []
        for response in responses:
            statuses = from_dict_list(Status, response.json())
            for status in statuses:
                lines = [separator] + list(status_lines(status))
                if len(batch_lines) + len(lines) > height - 2 and batch_lines:
                    yield "\n".join(batch_lines) + "\n" + separator
                    batch_lines = []
                batch_lines.extend(lines)

        if batch_lines:
            yield "\n".join(batch_lines) + "\n" + separator

    first = True
    printed_any = False
    for page in _page_generator():
        if not first and not get_continue():
            break
        if clear:
            click.clear()
        click.echo(page)
        first = False
        printed_any = True

    if not printed_any:
        click.echo("No statuses found")



import os
import sys

from pathlib import Path
from typing import Optional

from toot import App, User

CACHE_SUBFOLDER = "toot"


def save_last_post_id(app: App, user: User, id: str) -> None:
    """Save ID of the last post posted to this instance"""
    path = _last_post_id_path(app, user)
    with open(path, "w") as f:
        f.write(id)


def get_last_post_id(app: App, user: User) -> Optional[str]:
    """Retrieve ID of the last post posted to this instance"""
    path = _last_post_id_path(app, user)
    if path.exists():
        with open(path, "r") as f:
            return f.read()


def clear_last_post_id(app: App, user: User) -> None:
    """Delete the cached last post ID for this instance"""
    path = _last_post_id_path(app, user)
    path.unlink(missing_ok=True)


def _last_post_id_path(app: App, user: User):
    return get_cache_dir("last_post_ids") / f"{user.username}_{app.instance}"


def get_cache_dir(subdir: Optional[str] = None) -> Path:
    path = _cache_dir_path()
    if subdir:
        path = path / subdir
    path.mkdir(parents=True, exist_ok=True)
    return path


def _cache_dir_path() -> Path:
    """Returns the path to the cache directory"""

    # Windows
    if sys.platform == "win32" and "LOCALAPPDATA" in os.environ:
        return Path(os.environ["LOCALAPPDATA"], CACHE_SUBFOLDER)

    # Mac OS
    if sys.platform == "darwin":
        return Path.home() / "Library" / "Caches" / CACHE_SUBFOLDER

    # Respect XDG_CONFIG_HOME env variable if set
    # https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
    if "XDG_CACHE_HOME" in os.environ:
        return Path(os.environ["XDG_CACHE_HOME"], CACHE_SUBFOLDER)

    return Path.home() / ".cache" / CACHE_SUBFOLDER



import click
import re

from click import Context
from typing import Optional

from toot.cli import TUI_COLORS


def validate_language(ctx: Context, param: str, value: Optional[str]):
    if value is None:
        return None

    value = value.strip().lower()
    if re.match(r"^[a-z]{2}$", value):
        return value

    raise click.BadParameter("Language should be a two letter abbreviation.")


def validate_duration(ctx: Context, param: str, value: Optional[str]) -> Optional[int]:
    if value is None:
        return None

    match = re.match(r"""^
        (([0-9]+)\s*(days|day|d))?\s*
        (([0-9]+)\s*(hours|hour|h))?\s*
        (([0-9]+)\s*(minutes|minute|m))?\s*
        (([0-9]+)\s*(seconds|second|s))?\s*
    $""", value, re.X)

    if not match:
        raise click.BadParameter(f"Invalid duration: {value}")

    days = match.group(2)
    hours = match.group(5)
    minutes = match.group(8)
    seconds = match.group(11)

    days = int(match.group(2) or 0) * 60 * 60 * 24
    hours = int(match.group(5) or 0) * 60 * 60
    minutes = int(match.group(8) or 0) * 60
    seconds = int(match.group(11) or 0)

    duration = days + hours + minutes + seconds

    if duration == 0:
        raise click.BadParameter("Empty duration")

    return duration


def validate_instance(ctx: click.Context, param: str, value: Optional[str]):
    """
    Instance can be given either as a base URL or the domain name.
    Return the base URL.
    """
    if not value:
        return None

    value = value.rstrip("/")
    return value if value.startswith("http") else f"https://{value}"


def validate_tui_colors(ctx, param, value) -> Optional[int]:
    if value is None:
        return None

    if value in TUI_COLORS.values():
        return value

    if value in TUI_COLORS.keys():
        return TUI_COLORS[value]

    raise click.BadParameter(f"Invalid value: {value}. Expected one of: {', '.join(TUI_COLORS)}")


def validate_cache_size(ctx: click.Context, param: str, value: Optional[str]) -> Optional[int]:
    """validates the cache size parameter"""

    if value is None:
        return 1024 * 1024 * 10  # default 10MB
    else:
        if value.isdigit():
            size = int(value)
        else:
            raise click.BadParameter("Cache size must be numeric.")

    if size > 1024:
        raise click.BadParameter("Cache size too large: 1024MB maximum.")
    elif size < 1:
        raise click.BadParameter("Cache size too small: 1MB minimum.")
    return size


def validate_positive(_ctx: click.Context, _param: click.Parameter, value: Optional[int]):
    if value is not None and value <= 0:
        raise click.BadParameter("must be greater than 0")
    return value
